/**
 *      
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2006 by Bernd Breitenbach.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 */


/**
 * \file mazecairo.c
 * \author Bernd Breitenbach
 * \remarks
 *      bb - Aug 13, 2006: Created.
 */

#include <stdio.h>
#include <math.h>

#include <cairo.h>
/*
#ifdef WITH_CAIRO_PDF
#include <cairo-pdf.h>
#endif
#ifdef WITH_CAIRO_SVG
#include <cairo-svg.h>
#endif
*/
#include <cairo-svg.h>
#include <cairo-pdf.h>
#include "mazeintern.h"


static color_t ENTRANCE_COLOR ={0.8,0.,0.,1.};


static void _drawNumbers(cairo_t *cr,maze_2d_t *laby,render_parameter_t *para) 
{
  int i;
  double xsize= para->xcellsize;
  double ysize= para->ycellsize;
  cell_2d_t *cell;
  char buf[32];
  cairo_text_extents_t ex;

  cairo_set_source_rgba(cr,para->foreground.red,para->foreground.green,
			para->foreground.blue,para->foreground.alpha);
  for(i=0;i<MAZE_N_CELLS(laby);i++) {
    cell=MAZE_CELL(laby,i);
    snprintf(buf,31,"%d",CELL_NO(cell));
    cairo_text_extents(cr,buf,&ex);
    cairo_move_to(cr,CELL_X(cell)*xsize+para->xoff-ex.width/2.,
		  CELL_Y(cell)*ysize+para->yoff+ex.height/2.);
    cairo_show_text(cr,buf);
  }
}


static void _renderStroke(cairo_t *cr,stroke_t *stroke,render_parameter_t *para) 
{
  int i;
  for(i=0;i<STROKE_N_POINTS(stroke);i++) {
    if (STROKE_TYPE(stroke,i)==STROKE_MOVE) {
      cairo_move_to(cr,STROKE_X(stroke,i),STROKE_Y(stroke,i));
    } else {
      cairo_line_to(cr,STROKE_X(stroke,i),STROKE_Y(stroke,i));
    }
  }
  cairo_stroke(cr);
}


static void _renderText(cairo_t *cr,maze_2d_t *laby,stroke_t *stroke,render_parameter_t *para) 
{
  double xsize= para->xcellsize;
  double ysize= para->ycellsize;
  cell_2d_t *cell;
  cairo_text_extents_t ex;
  int i;
  char buf[2]="\0\0";

  cairo_set_source_rgba(cr,para->foreground.red,para->foreground.green,
			para->foreground.blue,para->foreground.alpha);
  for(i=0;i<MAZE_N_CELLS(laby);i++) {
    cell=MAZE_CELL(laby,i);
    if (CELL_USER_DATA(cell)>0) {
      buf[0]=(char)CELL_USER_DATA(cell);
      cairo_text_extents(cr,buf,&ex);
      cairo_move_to(cr,CELL_X(cell)*xsize+para->xoff-ex.width/2.,
		    CELL_Y(cell)*ysize+para->yoff+ex.height/2.);
      cairo_show_text(cr,buf);
    }
  }
}

static void _renderCopyright(cairo_t *cr,render_parameter_t *para) 
{
  cairo_text_extents_t ex;
  if (para->copyright_text!=NULL) {
    double red=para->background.red*0.8;
    double green=para->background.green*0.8;
    double blue=para->background.blue;
    double alpha=1.;//para->background.alpha;

    cairo_set_source_rgba(cr,red,green,blue,alpha);
    cairo_text_extents(cr,para->copyright_text,&ex);
    cairo_move_to(cr,para->width-ex.width-8,
		  para->height-ex.height/2.);
    cairo_show_text(cr,para->copyright_text);
  }
}


static void _renderMaze2D(cairo_surface_t *surface,maze_2d_t *laby,render_parameter_t *para) 
{
  cairo_t *cr=cairo_create(surface);
  stroke_t *stroke;

  cairo_set_line_cap(cr,CAIRO_LINE_CAP_ROUND);
  cairo_set_line_join(cr,CAIRO_LINE_JOIN_ROUND);
  cairo_set_line_width(cr,para->linewidth);

  if (!MAZE_TILEABLE(laby)) {
    cairo_set_source_rgba(cr,para->background.red,para->background.green,
			  para->background.blue,para->background.alpha);
    cairo_rectangle(cr,0,0,para->width,para->height);
    cairo_fill(cr);
  }
  _renderCopyright(cr,para);

  cairo_set_source_rgba(cr,para->foreground.red,para->foreground.green,
			para->foreground.blue,para->foreground.alpha);


  stroke=renderMaze2D(laby,para);
  _renderStroke(cr,stroke,para);
  freeStroke2D(stroke);

  if (para->draw_marks) {
    stroke=renderEntryExit(laby,MAZE_ENTRANCE(laby),para);
    _renderStroke(cr,stroke,para);
    freeStroke2D(stroke);

    stroke=renderEntryExit(laby,MAZE_EXIT(laby),para);
    _renderStroke(cr,stroke,para);
    freeStroke2D(stroke);
  }

  if (para->draw_solution || para->solution_text!=NULL) {
    cairo_set_source_rgba(cr,ENTRANCE_COLOR.red,
			  ENTRANCE_COLOR.green,
			  ENTRANCE_COLOR.blue,
			  ENTRANCE_COLOR.alpha);
    stroke=renderMaze2DSolution(laby,para);
    if (para->draw_solution) {
      _renderStroke(cr,stroke,para);
    }
    if (para->solution_text!=NULL) {
      renderTextAlongStroke(laby,stroke,para->solution_text);
      _renderText(cr,laby,stroke,para);
    }
    freeStroke2D(stroke);
  }

  
  if(para->drawno) {
    _drawNumbers(cr,laby,para);
  }

  cairo_show_page(cr);
  cairo_destroy(cr);
}



/**********************************************************************/

static void _setupSize(maze_2d_t *laby,render_parameter_t *para)
{
  maze_2d_dim_t dim;
  double cborder=0.;

  dim=maze2DDimension(laby);
  if (!MAZE_TILEABLE(laby)) {
    cborder=para->bordersize;
  }
  para->xoff=cborder-(para->xcellsize*dim.xmin);
  para->yoff=cborder-(para->ycellsize*dim.ymin);
  para->width =(dim.xmax-dim.xmin)*para->xcellsize+2*cborder;
  para->height=(dim.ymax-dim.ymin)*para->ycellsize+2*cborder;
}

/**********************************************************************/
static cairo_status_t  _writeToStdout(void *closure,const unsigned char *data,unsigned int length)
{
  int n=fwrite(data,length,1,stdout);
  return n==1 ? CAIRO_STATUS_SUCCESS : CAIRO_STATUS_WRITE_ERROR;
}


#ifdef WITH_CAIRO_PDF

static void _renderPDF2D(void (*func)(cairo_surface_t *,maze_2d_t *,render_parameter_t *),
			 maze_2d_t *laby,render_parameter_t *para)
{

  cairo_surface_t *surface=NULL;
  if (para->filename!=NULL) {
    surface=cairo_pdf_surface_create (para->filename,para->width,para->height);  
  } else {
    surface=cairo_pdf_surface_create_for_stream(&_writeToStdout,NULL,para->width,para->height);  
  }
  func(surface,laby,para);
  cairo_surface_finish(surface);
  cairo_surface_destroy(surface);
}

void renderPDF2D(maze_2d_t *laby,render_parameter_t *para)
{
  _setupSize(laby,para);
  _renderPDF2D(&_renderMaze2D,laby,para);
}

#endif

/**********************************************************************/

#ifdef WITH_CAIRO_SVG

static void _renderSVG2D(void (*func)(cairo_surface_t *,maze_2d_t *,render_parameter_t *),
			 maze_2d_t *laby,render_parameter_t *para)
{
  cairo_surface_t *surface=NULL;
  if (para->filename!=NULL) {
    surface=cairo_svg_surface_create (para->filename,para->width,para->height);  
  } else {
    surface=cairo_svg_surface_create_for_stream(&_writeToStdout,NULL,para->width,para->height);  
  }
  func(surface,laby,para);
  cairo_surface_destroy(surface);
}

void renderSVG2D(maze_2d_t *laby,render_parameter_t *para)
{
  _setupSize(laby,para);
  _renderSVG2D(&_renderMaze2D,laby,para);
}

#endif

/**********************************************************************/

#ifdef WITH_CAIRO_IMAGE

static void _renderPNG2D(void (*func)(cairo_surface_t *,maze_2d_t *,render_parameter_t *),
			 maze_2d_t *laby,render_parameter_t *para)
{
  cairo_surface_t *surface=cairo_image_surface_create (CAIRO_FORMAT_ARGB32,para->width,para->height);
  func(surface,laby,para);
  if (para->filename!=NULL) {
    cairo_surface_write_to_png(surface,para->filename);
  } else {
    cairo_surface_write_to_png_stream(surface,&_writeToStdout,NULL);
  }
  cairo_surface_finish(surface);
  cairo_surface_destroy(surface);
}


void renderPNG2D(maze_2d_t *laby,render_parameter_t *para)
{
  _setupSize(laby,para);
  _renderPNG2D(&_renderMaze2D,laby,para);
}

#endif

/**********************************************************************/

maze_2d_mask_t *newMaze2DMask(char *filename)
{
  cairo_surface_t *surface=cairo_image_surface_create_from_png(filename);
  maze_2d_mask_t *mask=NULL;
  if (cairo_surface_status(surface)!=CAIRO_STATUS_SUCCESS) {
    fprintf(stderr,"%s: %s\n",filename,cairo_status_to_string(cairo_surface_status(surface)));
  } else {
    cairo_format_t format=cairo_image_surface_get_format(surface);
    if (format==CAIRO_FORMAT_ARGB32 || format==CAIRO_FORMAT_RGB24) {
      mask=(maze_2d_mask_t*)surface;
    } else {
      fprintf(stderr,"%s: unsupported image format\n",filename);
    }
  }
  return mask;
}

void freeMaze2DMask(maze_2d_mask_t *mask)
{
  cairo_surface_t *surface=(cairo_surface_t *)mask;
  cairo_surface_finish(surface);
  cairo_surface_destroy(surface);
}


int maze2DMaskQuery(maze_2d_mask_t *mask,double x,double y)
{
  int retVal=TRUE;
  if (mask!=NULL) {
    cairo_surface_t *surface=(cairo_surface_t *)mask;
    int w=cairo_image_surface_get_width(surface);
    int h=cairo_image_surface_get_height(surface);
    unsigned char *data=cairo_image_surface_get_data(surface);
    int stride=cairo_image_surface_get_stride(surface);
    int mx,my;

    if (x<0.) x=0.;
    else if (x>1) x=1.;
    if (y<0.) y=0.;
    else if (y>1) y=1.;
    mx=(int)(x*w);
    my=(int)(y*h);
    data+=stride*my+mx*4;
    retVal=data[0]>0x80 && data[1]>0x80 && data[2]>0x80;
  }
  return retVal;
}

/**********************************************************************/



