/**
 *      
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2006 by Bernd Breitenbach.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 */

/**
 * \file mazelib.c
 * \brief 
 * \author Bernd Breitenbach
 * \remarks
 *      bb - Aug 10, 2006: Created.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <ctype.h>

#include "mazeintern.h"


/**********************************************************************/

static void _setupOff(maze_2d_t *maze)
{
  cell_2d_t *_off=MAZE_OFF(maze);
  CELL_TYPE(_off)=TYPE_OFF;
  CELL_STATE(_off)=CELL_OUT;
  visitCell(_off);
  CELL_NO(_off)=-1;
  CELL_X(_off)=COORD_OFF;
  CELL_Y(_off)=COORD_OFF;
  CELL_Z(_off)=COORD_OFF;
  CELL_N_NEIGHBOURS(_off)=0;
  CELL_NEIGHBOURS(_off)=NULL;
}

/**********************************************************************/


static void freeCell2D(cell_2d_t *cell) 
{
  if (CELL_NEIGHBOURS(cell)!=NULL) {
    free(CELL_NEIGHBOURS(cell));
  }
}

void freeMaze2D(maze_2d_t *maze) {
  int i;
  for(i=0;i<MAZE_COLUMNS(maze)*MAZE_ROWS(maze);i++) {
    freeCell2D(&maze->cells[i]);
  }
  MAZE_COLUMNS(maze)=0;
  MAZE_ROWS(maze)=0;
  free(maze->cells);
  free(maze);
}

/**********************************************************************/

path_2d_t *newPath2D(int steps)
{
  path_2d_t *path=ALLOC(path_2d_t);
  PATH_N_STEPS(path)=steps;
  PATH_STEPS(path)=CALLOC(steps,point_2d_t);
  return path;
}


void freePath2D(path_2d_t *path)
{
  free(PATH_STEPS(path));
  free(path);
}


/**********************************************************************/

cell_2d_t *cell2D(maze_2d_t *maze,int col, int row) 
{
  cell_2d_t *cell=MAZE_OFF(maze);
  if (col>=0 && col<MAZE_COLUMNS(maze) && row>=0 && row<MAZE_ROWS(maze)) {
    cell=&maze->cells[row*MAZE_COLUMNS(maze)+col];
  } else if (MAZE_TILEABLE(maze)) {
    row=(row+MAZE_ROWS(maze))%MAZE_ROWS(maze);
    col=(col+MAZE_COLUMNS(maze))%MAZE_COLUMNS(maze);
    cell=&maze->cells[row*MAZE_COLUMNS(maze)+col];
  }
  return cell;
}


maze_2d_t *new2DMaze(maze_parameter_t *parameter)
{
  maze_2d_t *maze=ALLOC(maze_2d_t);
  int x,y,i,j,n=0;
  div_t dx,dy;
  cell_2d_t *cur,*neighbour;
  cell_type_t kind;
  cell_structure_t *structure=parameter->structure;
  

  MAZE_STRUCT(maze)=structure;
  if (parameter->tileable) {
    n=parameter->rows%STRUCT_ROWS(structure);
    if (n!=0) {
      parameter->rows+=(STRUCT_ROWS(structure)-n);
      INFO("increasing no of rows to %d\n",parameter->rows);
    }
    n=parameter->columns%STRUCT_COLUMNS(structure);
    if (n!=0) {
      parameter->columns+=(STRUCT_COLUMNS(structure)-n);
      INFO("increasing no of columns to %d\n",parameter->columns);
    }
  }
  MAZE_COLUMNS(maze)=parameter->columns;
  MAZE_ROWS(maze)=parameter->rows;
  MAZE_TILEABLE(maze)=parameter->tileable;
  MAZE_TYPE(maze)=MAZE_POLY_2D;
  MAZE_N_CELLS(maze)=MAZE_ROWS(maze)*MAZE_COLUMNS(maze);
  MAZE_CELLS(maze)=CALLOC(MAZE_N_CELLS(maze)+1,cell_2d_t);

  _setupOff(maze);

  for(y=0;y<MAZE_ROWS(maze);y++) {
    dy=div(y,STRUCT_ROWS(structure));
    for(x=0;x<MAZE_COLUMNS(maze);x++) {
      dx=div(x,STRUCT_COLUMNS(structure));
      kind=STRUCT_CELL_TYPE(structure,dx.rem,dy.rem);
      cur=cell2D(maze,x,y);
      CELL_TYPE(cur)=kind;
      if (kind!=TYPE_NONE) {
	CELL_BORDER(cur)=STRUCT_BORDER(structure,kind);
	CELL_NO(cur)=n++;
	CELL_X(cur)=(double)dx.quot*STRUCT_WIDTH(structure) 
	  +STRUCT_CENTER_X(structure,dx.rem,dy.rem);
	CELL_Y(cur)=(double)dy.quot*STRUCT_HEIGHT(structure)
	  +STRUCT_CENTER_Y(structure,dx.rem,dy.rem);
	CELL_Z(cur)=COORD_ZERO;
	CELL_N_NEIGHBOURS(cur)=STRUCT_N_NEIGHBOURS(structure,kind);
	CELL_NEIGHBOURS(cur)=CALLOC(CELL_N_NEIGHBOURS(cur),cell_2d_conn_t);
	for(i=0;i<CELL_N_NEIGHBOURS(cur);i++) {
	  CELL_NEIGHBOUR(cur,i)=cell2D(maze,x+STRUCT_DX(structure,kind,i),
				       y+STRUCT_DY(structure,kind,i));
	}
      }
    }
  }
  for(y=0;y<MAZE_ROWS(maze);y++) {
    for(x=0;x<MAZE_COLUMNS(maze);x++) {
      cur=cell2D(maze,x,y);
      if (CELL_TYPE(cur)!=TYPE_NONE) {
	if (!maze2DMaskQuery(parameter->mask,
			     x/(double)MAZE_COLUMNS(maze),
			     y/(double)MAZE_ROWS(maze))) {
	  CELL_TYPE(cur)=TYPE_OFF;
	} else {
	}
	for(i=0;i<CELL_N_NEIGHBOURS(cur);i++) {
	  CELL_WALL(cur,i)=WALL_CLOSED;
	  CELL_NEIGHBOUR_IDX(cur,i)=-1;
	  neighbour=CELL_NEIGHBOUR(cur,i);
	  for(j=0;j<CELL_N_NEIGHBOURS(neighbour);j++) {
	    if (CELL_NEIGHBOUR(neighbour,j)==cur) {
	      CELL_NEIGHBOUR_IDX(cur,i)=j;
	      break;
	    }
	  }
	}
      }
    }
  }
  return maze;  
}

/**********************************************************************/

maze_2d_dim_t maze2DDimension(maze_2d_t *maze)
{
  int y,x,j;
  cell_2d_t *cur;
  maze_2d_dim_t dim={.xmin=9E9,.ymin=9E9,.xmax=-9E9,.ymax=-9E9};
  cell_structure_t *s=MAZE_STRUCT(maze);
  int xlim=MAZE_COLUMNS(maze);
  int ylim=MAZE_ROWS(maze);
  cell_type_t type;

  for(y=0;y<ylim;y++) {
    for(x=0;x<xlim;x++) {
      cur=cell2D(maze,x,y);
      if (CELL_VALID(cur)) {
	if (!MAZE_TILEABLE(maze)) {
	  type=CELL_TYPE(cur);
	  for(j=0;j<CELL_N_NEIGHBOURS(cur);j++) {
	    dim.xmin=MIN(CELL_X(cur)+STRUCT_PX(s,type,j),dim.xmin);
	  }
	} else {
	  dim.xmin=MIN(CELL_X(cur),dim.xmin);
	}
	break;
      }
    }

    for(x=xlim-1;x>=0;x--) {
      cur=cell2D(maze,x,y);
      if (CELL_VALID(cur)) {
	if (!MAZE_TILEABLE(maze)) {
	  type=CELL_TYPE(cur);
	  for(j=0;j<CELL_N_NEIGHBOURS(cur);j++) {
	    dim.xmax=MAX(CELL_X(cur)+STRUCT_PX(s,type,j),dim.xmax);
	  }
	} else {
	  dim.xmax=MAX(CELL_X(cur),dim.xmax);
	}
	break;
      }
    }
  }

  for(x=0;x<xlim;x++) {
    for(y=0;y<ylim;y++) {
      cur=cell2D(maze,x,y);
      if (CELL_VALID(cur)) {
	if (!MAZE_TILEABLE(maze)) {
	  type=CELL_TYPE(cur);
	  for(j=0;j<CELL_N_NEIGHBOURS(cur);j++) {
	    dim.ymin=MIN(CELL_Y(cur)+STRUCT_PY(s,type,j),dim.ymin);
	  }
	} else {
	  dim.ymin=MIN(CELL_Y(cur),dim.ymin);
	}
	break;
      }
    }

    for(y=ylim-1;y>=0;y--) {
      cur=cell2D(maze,x,y);
      if (CELL_VALID(cur)) {
	type=CELL_TYPE(cur);
	if (!MAZE_TILEABLE(maze)) {
	  for(j=0;j<CELL_N_NEIGHBOURS(cur);j++) {
	    dim.ymax=MAX(CELL_Y(cur)+STRUCT_PY(s,type,j),dim.ymax);
	  }
	} else {
	  dim.ymax=MAX(CELL_Y(cur),dim.ymax);
	}
	break;
      }
    }
  }
  if (MAZE_TILEABLE(maze)) {
    dim.xmin+=STRUCT_XMINADD(s);
    dim.xmax+=STRUCT_XMAXADD(s);
    dim.ymin+=STRUCT_YMINADD(s);
    dim.ymax+=STRUCT_YMAXADD(s);
  }
  return dim;
}


/**********************************************************************/

static int _directNeighbours(cell_2d_t *cell1,int idx1,cell_2d_t *cell2,int idx2)
{
  double dx1=fabs(CELL_X(cell1)-CELL_X(cell2));
  double dy1=fabs(CELL_Y(cell1)-CELL_Y(cell2));

  double dx2=fabs(CELL_BORDER_X(cell1,idx1)-CELL_BORDER_X(cell2,idx2));
  double dy2=fabs(CELL_BORDER_Y(cell1,idx1)-CELL_BORDER_Y(cell2,idx2));

  return dx1*dx1+dy1*dy1<2*(dx2*dx2+dy2*dy2);
}

void resetCell2DUserData(cell_2d_t *cell)
{
  int j;
  CELL_USER_DATA(cell)=0;
  for(j=0;j<CELL_N_NEIGHBOURS(cell);j++) {
    CELL_NEIGHBOUR_DATA(cell,j)=0;
  }
}

void resetMaze2DUserData(maze_2d_t *maze)
{
  cell_2d_t *cur;
  int i,j;

  for(i=0;i<MAZE_N_CELLS(maze);i++) {
    cur=MAZE_CELL(maze,i);
    CELL_USER_DATA(cur)=0;
    for(j=0;j<CELL_N_NEIGHBOURS(cur);j++) {
      CELL_NEIGHBOUR_DATA(cur,j)=0;
    }
  }
}


int visitCell(cell_2d_t *cell) 
{
  int visit=FALSE;
  if (CELL_VALID(cell) &&  CELL_USER_DATA(cell)==NOT_VISITED) {
    CELL_USER_DATA(cell)=VISITED;
    visit=TRUE;
  }
  return visit;
}


int visitNeighbour(cell_2d_t *cell,int idx) 
{
  return visitCell(CELL_NEIGHBOUR(cell,idx));
}


int visitWall(cell_2d_t *cell,int idx) 
{
  int visit=FALSE;
  if (CELL_NEIGHBOUR_DATA(cell,idx)==NOT_VISITED) {
    CELL_NEIGHBOUR_DATA(cell,idx)=VISITED;
    int nidx=CELL_NEIGHBOUR_IDX(cell,idx);
    if (nidx>=0) {
      cell_2d_t *neighbour=CELL_NEIGHBOUR(cell,idx);
      if (1 || _directNeighbours(cell,idx,neighbour,nidx)) {
	CELL_NEIGHBOUR_DATA(CELL_NEIGHBOUR(cell,idx),nidx)=VISITED;
      }
    }
    visit=TRUE;
  }
  return visit;
}

int wallVisited(cell_2d_t *cell,int idx) 
{
  return CELL_NEIGHBOUR_DATA(cell,idx)==VISITED;
}

/**********************************************************************/

static void _setWall2D(cell_2d_t *cell,int idx,wall_type_t wall) 
{
  if (CELL_VALID(cell)) {
    CELL_WALL(cell,idx)=wall;
  }
}

void setWall2D(cell_2d_t *cell,int idx,wall_type_t wall) 
{
  _setWall2D(cell,idx,wall);
  _setWall2D(CELL_NEIGHBOUR(cell,idx),CELL_NEIGHBOUR_IDX(cell,idx),wall);
}

void makeOpening2D(cell_2d_t *cell,int idx) 
{
  setWall2D(cell,idx,WALL_OPEN);
}


/**********************************************************************/


wall_type_t wall2D(cell_2d_t *cell,int idx) 
{
  return CELL_WALL(cell,idx);
}


int isOpen2D(cell_2d_t *cell,int idx) 
{
  return wall2D(cell,idx)==WALL_OPEN;
}

int isClosed2D(cell_2d_t *cell,int idx) 
{
  return wall2D(cell,idx)==WALL_CLOSED;
}


static int _carveOut2D(maze_2d_t *maze,cell_2d_t *cell) 
{
  int i;
  int n=CELL_N_NEIGHBOURS(cell);
  for(i=0;i<n;i++) {
    if (CELL_TYPE(CELL_NEIGHBOUR(cell,i))==TYPE_OFF) {
      makeOpening2D(cell,i);
      return TRUE;
    }
  }
  return FALSE;
}

/**********************************************************************/

static cell_2d_t *_randomBegin(maze_2d_t *maze,void *data)
{
  int n=rand()%MAZE_N_CELLS(maze);
  cell_2d_t *cell;
  cell=MAZE_CELL(maze,n);
  while(!CELL_VALID(cell)) {
    n=(n+1)%MAZE_N_CELLS(maze);
    cell=MAZE_CELL(maze,n);
  }
  MAZE_ENTRANCE(maze)=cell;
  return cell;
}

/**********************************************************************/

struct _random_carver {
  carver_2d_t    carver;
  int            (*direction)(cell_2d_t *,void*);
  double         windingness;
  int            last_direction;
};


static void _step2D(struct _random_carver *carver,cell_2d_t *cell)
{
  int i;
  int idx=CARVER_DIRECTION(carver)(cell,carver); 
  
  for(i=0;i<CELL_N_NEIGHBOURS(cell);i++) {
    cell_2d_t *neighbour=CELL_NEIGHBOUR(cell,idx);
    if (visitCell(neighbour)) {
      makeOpening2D(cell,idx);
      _step2D(carver,neighbour);
    }
    idx++;
  }
}

static void _recursiveBacktracker2D(carver_2d_t *carver)
{
  resetMaze2DUserData(CARVER_LABY(carver));
  cell_2d_t *cell=_randomBegin(CARVER_LABY(carver),carver);
  if (visitCell(cell)) {
    _step2D((struct _random_carver *)carver,cell);
  }
}

/**********************************************************************/

typedef struct _ptr_collection
{
  int             n_items;
  int             max_items;
  void            **items;
} ptr_collection_t;

ptr_collection_t *newCollection(int size)
{
  ptr_collection_t *collection=CALLOC(1,ptr_collection_t);
  COLL_MAX_ITEMS(collection)=size;
  COLL_N_ITEMS(collection)=0;
  collection->items=CALLOC(size,void *);
  return collection;
}

void freeCollection(ptr_collection_t *collection)
{
  free(collection->items);
  free(collection);
}

void resizeCollection(ptr_collection_t *collection,int new_size)
{
  COLL_MAX_ITEMS(collection)=new_size;
  collection->items=REALLOC(collection->items,new_size,void *);
}

void collectionAdd(ptr_collection_t *collection,void *item)
{
  if (COLL_N_ITEMS(collection)==COLL_MAX_ITEMS(collection)) {
    resizeCollection(collection,2*COLL_MAX_ITEMS(collection));
  }
  collection->items[COLL_N_ITEMS(collection)]=item;
  COLL_N_ITEMS(collection)+=1;
}

void *collectionRemove(ptr_collection_t *collection,int idx)
{
  void *retVal=COLL_ITEM(collection,idx,void *);
  collection->items[idx]=collection->items[COLL_N_ITEMS(collection)-1];
  COLL_N_ITEMS(collection)-=1;
  return retVal;
}

/**********************************************************************/

static void _primCarver2D(carver_2d_t *carver)
{
  int i,idx,n;
  maze_2d_t *maze=CARVER_LABY(carver);
  resetMaze2DUserData(CARVER_LABY(carver));
  cell_2d_t *cur=_randomBegin(maze,carver);
  cell_2d_t *neighbour;
  ptr_collection_t *coll=newCollection(MAZE_N_CELLS(maze));
  while (TRUE) {
    CELL_USER_DATA(cur)=VISITED;
    for(i=0;i<CELL_N_NEIGHBOURS(cur);i++) {
      neighbour=CELL_NEIGHBOUR(cur,i);
      if (CELL_VALID(neighbour) && CELL_USER_DATA(neighbour)==NOT_VISITED) {
	CELL_USER_DATA(neighbour)=FRONTIER;
	collectionAdd(coll,neighbour);
      }
    }
    if(COLL_N_ITEMS(coll)==0) {
      break;
    }
    idx=rand()%COLL_N_ITEMS(coll);
    cur=COLL_ITEM(coll,idx,cell_2d_t*);
    collectionRemove(coll,idx);
    n=rand()%CELL_N_NEIGHBOURS(cur);
    for(i=0;i<CELL_N_NEIGHBOURS(cur);i++) {
      neighbour=CELL_NEIGHBOUR(cur,n);
      if (CELL_VALID(neighbour) && CELL_USER_DATA(neighbour)==VISITED) {
	makeOpening2D(cur,n);
	break;
      }
      n++;
    }    
  } 
  freeCollection(coll);
}


/**********************************************************************/

static int _randomDirection(cell_2d_t *cell,void *data)
{
  struct _random_carver *carver=(struct _random_carver*)data;
  int n=CELL_N_NEIGHBOURS(cell);
  double fac=carver->windingness*( (rand()/(double)RAND_MAX) - 0.5);
  if (fac<-0.5) fac=-0.5;
  else if (fac>0.5) fac=0.5;
  if (fac<0.) fac=1+fac;
  carver->last_direction=(carver->last_direction+(int)(n*fac))%n;
  return carver->last_direction;
}

/**********************************************************************/


struct _path_data
{
  carver_2d_t *carver;
  int len;
  cell_2d_t *farest;
  cell_predicate_f predicate;
};


static int _True(cell_2d_t *cell) 
{
  return TRUE;
}


static int _atBorder(cell_2d_t *cell) 
{
  int i;
  for(i=0;i<CELL_N_NEIGHBOURS(cell);i++) {
    if (CELL_TYPE(CELL_NEIGHBOUR(cell,i))==TYPE_OFF) {
      return TRUE;
    }
  }
  return FALSE;
}

static int _atNorth(cell_2d_t *cell) 
{
  int i;
  for(i=0;i<CELL_N_NEIGHBOURS(cell);i++) {
    if (CELL_TYPE(CELL_NEIGHBOUR(cell,i))==TYPE_OFF &&
	CELL_NEIGHBOUR_DY(cell,i)<0 && CELL_NEIGHBOUR_DX(cell,i)==0) {
      return TRUE;
    }
  }
  return FALSE;
}

static int _atSouth(cell_2d_t *cell) 
{
  int i;
  for(i=0;i<CELL_N_NEIGHBOURS(cell);i++) {
    if (CELL_TYPE(CELL_NEIGHBOUR(cell,i))==TYPE_OFF &&
	CELL_NEIGHBOUR_DY(cell,i)>0 && CELL_NEIGHBOUR_DX(cell,i)==0) {
      return TRUE;
    }
  }
  return FALSE;
}

static int _atWest(cell_2d_t *cell) 
{
  int i;
  for(i=0;i<CELL_N_NEIGHBOURS(cell);i++) {
    if (CELL_TYPE(CELL_NEIGHBOUR(cell,i))==TYPE_OFF &&
	CELL_NEIGHBOUR_DX(cell,i)<0 && CELL_NEIGHBOUR_DY(cell,i)==0) {
      return TRUE;
    }
  }
  return FALSE;
}

static int _atEast(cell_2d_t *cell) 
{
  int i;
  for(i=0;i<CELL_N_NEIGHBOURS(cell);i++) {
    if (CELL_TYPE(CELL_NEIGHBOUR(cell,i))==TYPE_OFF &&
	CELL_NEIGHBOUR_DX(cell,i)>0 && CELL_NEIGHBOUR_DY(cell,i)==0)  {
      return TRUE;
    }
  }
  return FALSE;
}



static void _findFarest(cell_2d_t *cell,int idx,struct _path_data *data) 
{
  int i;
  cell_2d_t *neighbour=NULL;

  if (idx>data->len && data->predicate(cell)) {
    data->len=idx;
    data->farest=cell;
  }

  for(i=0;i<CELL_N_NEIGHBOURS(cell);i++) {
    neighbour=CELL_NEIGHBOUR(cell,i);
    if (isOpen2D(cell,i) && visitCell(neighbour)) {
      _findFarest(neighbour,idx+1,data);
    }
  }
}

/**********************************************************************/

static void _setupPredicate(struct _path_data *data,direction_t dir,int tileable)
{
  if (tileable) {
    data->predicate=&_True;
  } else {
    switch(dir) {
    case NORTH:
      data->predicate=&_atNorth;
      break;
    case WEST:
      data->predicate=&_atWest;
      break;
    case SOUTH:
      data->predicate=&_atSouth;
      break;
    case EAST:
      data->predicate=&_atEast;
      break;
    default:
      data->predicate=&_atBorder;
      break;
    }
  }
}

static void _searchEntrance(maze_2d_t *maze,carver_2d_t *carver) 
{
  resetMaze2DUserData(maze);

  struct _path_data data={carver,0,NULL,NULL};

  _setupPredicate(&data,CARVER_PARA(carver)->entrance,CARVER_PARA(carver)->tileable);
  _findFarest(MAZE_ENTRANCE(maze),0,&data);
  MAZE_ENTRANCE(maze)=data.farest;
  _carveOut2D(maze,MAZE_ENTRANCE(maze));
}

static void _searchExit(maze_2d_t *maze,carver_2d_t *carver) 
{
  resetMaze2DUserData(maze);

  struct _path_data data={carver,0,NULL,NULL};
  
  _setupPredicate(&data,CARVER_PARA(carver)->exit,CARVER_PARA(carver)->tileable);
  _findFarest(MAZE_ENTRANCE(maze),0,&data);
  MAZE_EXIT(maze)=data.farest;
  _carveOut2D(maze,MAZE_EXIT(maze));
}


/**********************************************************************/

carver_2d_t *newRandomBacktracker2D(maze_parameter_t *parameter) 
{
  struct _random_carver *randomCarver=ALLOC(struct _random_carver);
  if (parameter->seed<0) {
    parameter->seed=(int)time(NULL);
  }
  srand(parameter->seed);
  randomCarver->windingness=1.0;
  randomCarver->last_direction=0;
  randomCarver->windingness=pow(4,parameter->windingness)-1;
  randomCarver->windingness=randomCarver->windingness*randomCarver->windingness;

  carver_2d_t *carver=(carver_2d_t*)randomCarver;

  CARVER_TYPE(carver)=CARVER_BACKTRACKER;
  CARVER_PARA(carver)    =parameter;
  CARVER_CARVE(carver)   =&_recursiveBacktracker2D;
  randomCarver->direction=&_randomDirection;
  CARVER_ENTRANCE(carver)=&_searchEntrance;
  CARVER_EXIT(carver)    =&_searchExit;
  return (carver_2d_t*)randomCarver;
}

/**********************************************************************/

carver_2d_t *newPrimCarver2D(maze_parameter_t *parameter) 
{
  carver_2d_t *carver=ALLOC(carver_2d_t);
  if (parameter->seed<0) {
    parameter->seed=(int)time(NULL);
  }
  srand(parameter->seed);

  CARVER_TYPE(carver)=CARVER_PRIM;
  CARVER_PARA(carver)    =parameter;
  CARVER_CARVE(carver)   =&_primCarver2D;
  CARVER_ENTRANCE(carver)=&_searchEntrance;
  CARVER_EXIT(carver)    =&_searchExit;
  
  return (carver_2d_t*)carver;
}

/**********************************************************************/

void freeCarver2D(carver_2d_t *carver)
{
  free(carver);
}

/**********************************************************************/

void carveMaze2D(carver_2d_t *carver,maze_2d_t *maze)
{
  resetMaze2DUserData(maze);
  CARVER_LABY(carver)=maze;
  CARVER_CARVE(carver)(carver);
  CARVER_ENTRANCE(carver)(maze,carver);
  CARVER_EXIT(carver)(maze,carver);
}

/**********************************************************************/

static stroke_t *_appendSubStrokes(stroke_t *stroke,double x1,double y1,double x2,double y2,
				   render_parameter_t *para) 
{
  int i,n;
  double dx,dy;
  dx=x2-x1;
  dy=y2-y1;
  double scale;
  n=para->n_fuzzy;
  for(i=0;i<n;i++) {
    scale=(0.5-(rand()/(double)RAND_MAX))*para->amplitude;
    stroke=appendStroke2DPoint(stroke,STROKE_LINE,
			       x1+i/(double)n*dx+scale*dy,
			       y1+i/(double)n*dy-scale*dx);
  }
  return stroke;
}

static stroke_t *_renderCell2D(maze_2d_t *maze,cell_2d_t *cell,stroke_t *stroke,render_parameter_t *para) 
{
  int j,n,cn=0;
  int line_mode=FALSE;
  cell_type_t type;
  double x1,y1,x2,y2;
  cell_structure_t *s=MAZE_STRUCT(maze);
  double xsize= para->xcellsize;
  double ysize= para->ycellsize;

  n=CELL_N_NEIGHBOURS(cell);
  for(j=0;j<n;j++) {
    if (!isClosed2D(cell,j)) {
      cn=j;
      break;
    }
  }      
  for(j=0;j<n;j++) {
    if (isClosed2D(cell,cn) && visitWall(cell,cn)) {
      type=CELL_TYPE(cell);
      if (!line_mode) {
	line_mode=TRUE;
	x1=(CELL_X(cell)+STRUCT_PX(s,type,cn))*xsize+para->xoff;
	y1=(CELL_Y(cell)+STRUCT_PY(s,type,cn))*ysize+para->yoff;
	stroke=appendStroke2DPoint(stroke,STROKE_MOVE,x1,y1);
      } 
      x2=(CELL_X(cell)+STRUCT_PX(s,type,cn+1))*xsize+para->xoff;
      y2=(CELL_Y(cell)+STRUCT_PY(s,type,cn+1))*ysize+para->yoff;
      if (para->fuzzy) {
	stroke=_appendSubStrokes(stroke,x1,y1,x2,y2,para);
      }
      stroke=appendStroke2DPoint(stroke,STROKE_LINE,x2,y2);
      x1=x2;
      y1=y2;
    } else {
      line_mode=FALSE;
    }
    cn=(cn+1)%n;
  }
  return stroke;
}


stroke_t *renderMaze2D(maze_2d_t *maze,render_parameter_t *para) 
{
  int i;
  cell_2d_t *cell;
  stroke_t *stroke=newStroke2D(3*MAZE_N_CELLS(maze));
  double xoff,yoff,xtmp,ytmp;
  cell_structure_t *s=MAZE_STRUCT(maze);

  resetMaze2DUserData(maze);
  for(i=0;i<MAZE_N_CELLS(maze);i++) {
    cell=MAZE_CELL(maze,i);
    if (CELL_VALID(cell)) {
      stroke=_renderCell2D(maze,cell,stroke,para);
    }
  }
  if (MAZE_TILEABLE(maze)) {
    xoff=MAZE_COLUMNS(maze)/STRUCT_COLUMNS(s)*STRUCT_WIDTH(s)*para->xcellsize;
    yoff=MAZE_ROWS(maze)/STRUCT_ROWS(s)*STRUCT_HEIGHT(s)*para->ycellsize;
    resetMaze2DUserData(maze);
    xtmp=para->xoff;
    ytmp=para->yoff;

    para->xoff+=xoff;
    for (i=0;i<MAZE_ROWS(maze);i++) {
      resetCell2DUserData(cell2D(maze,MAZE_COLUMNS(maze),i));
    }
    for (i=0;i<MAZE_ROWS(maze);i++) {
      cell=cell2D(maze,MAZE_COLUMNS(maze),i);
      stroke=_renderCell2D(maze,cell,stroke,para);
    }
    para->xoff=xtmp;

    para->yoff+=yoff;
    for (i=0;i<MAZE_COLUMNS(maze);i++) {
      resetCell2DUserData(cell2D(maze,i,MAZE_ROWS(maze)));
    }
    for (i=0;i<MAZE_COLUMNS(maze);i++) {
      cell=cell2D(maze,i,MAZE_ROWS(maze));
      stroke=_renderCell2D(maze,cell,stroke,para);
    }
    para->xoff+=xoff;
    resetCell2DUserData(cell2D(maze,MAZE_COLUMNS(maze),MAZE_ROWS(maze)));
    stroke=_renderCell2D(maze,cell2D(maze,MAZE_COLUMNS(maze),MAZE_ROWS(maze)),stroke,para);

    para->xoff=xtmp;
    para->yoff=ytmp;
  }
  return stroke;
}



stroke_t *renderEntryExit(maze_2d_t *maze,cell_2d_t *cell,render_parameter_t *para) 
{
  int i;
  cell_structure_t *s=MAZE_STRUCT(maze);
  double xsize= para->xcellsize;
  double ysize= para->ycellsize;
  double x1,y1,x2,y2;
  double dx,dy,xm,ym,l,dx1,dx2,dy1,dy2;
  cell_type_t type;
  stroke_t *stroke=newStroke2D(6);

  if (cell) {
    type=CELL_TYPE(cell);
    for(i=0;i<CELL_N_NEIGHBOURS(cell);i++) {
      if (isOpen2D(cell,i)) {
	cell_2d_t *neighbour=CELL_NEIGHBOUR(cell,i);
	if (CELL_TYPE(neighbour)==TYPE_OFF) {
	  xm=(STRUCT_PX(s,type,i+1)+STRUCT_PX(s,type,i))*0.5;
	  ym=(STRUCT_PY(s,type,i+1)+STRUCT_PY(s,type,i))*0.5;
	  dx=STRUCT_PX(s,type,i+1)-STRUCT_PX(s,type,i);
	  dy=STRUCT_PY(s,type,i+1)-STRUCT_PY(s,type,i);
	  l=sqrt(dx*dx+dy*dy);
	  dx/=l;
	  dy/=l;
	  x1=(xm+CELL_X(cell)-dy*0.25)*xsize+para->xoff;
	  y1=(ym+CELL_Y(cell)+dx*0.25)*ysize+para->yoff;
	  x2=(xm+CELL_X(cell)+dy*0.25)*xsize+para->xoff;
	  y2=(ym+CELL_Y(cell)-dx*0.25)*ysize+para->yoff;
	  dx1=(CELL_X(cell)*xsize+para->xoff-x1);
	  dx2=(CELL_X(cell)*xsize+para->xoff-x2);
	  dy1=(CELL_Y(cell)*ysize+para->yoff-y1);
	  dy2=(CELL_Y(cell)*ysize+para->yoff-y2);
	  if ( (dx1*dx1+dy1*dy1<dx2*dx2+dy2*dy2) ^ (cell==MAZE_ENTRANCE(maze)) ) {
	    stroke=appendStroke2DPoint(stroke,STROKE_MOVE,(x1+x2)*0.5+(y2-y1)*0.3,(y1+y2)*0.5-(x2-x1)*0.3);
	    stroke=appendStroke2DPoint(stroke,STROKE_LINE,x2,y2);
	    stroke=appendStroke2DPoint(stroke,STROKE_LINE,(x1+x2)*0.5-(y2-y1)*0.3,(y1+y2)*0.5+(x2-x1)*0.3);
	  } else {
	    stroke=appendStroke2DPoint(stroke,STROKE_MOVE,(x1+x2)*0.5+(y1-y2)*0.3,(y1+y2)*0.5-(x1-x2)*0.3);
	    stroke=appendStroke2DPoint(stroke,STROKE_LINE,x1,y1);
	    stroke=appendStroke2DPoint(stroke,STROKE_LINE,(x1+x2)*0.5-(y1-y2)*0.3,(y1+y2)*0.5+(x1-x2)*0.3);
	  }
	  stroke=appendStroke2DPoint(stroke,STROKE_MOVE,x1,y1);
	  stroke=appendStroke2DPoint(stroke,STROKE_LINE,x2,y2);
	  break;
	}
      }
    }
  }
  return stroke;
}


struct _solution
{
  stroke_t           *stroke;
  cell_2d_t          *exit;
  render_parameter_t *para;
};

static int _seperatedCells(cell_2d_t *cell1,cell_2d_t *cell2) 
{
  double dx=CELL_X(cell1)-CELL_X(cell2);
  double dy=CELL_Y(cell1)-CELL_Y(cell2);
  if (dx*dx+dy*dy>4) {
    printf("dx:%g dy:%g\n",dx,dy);
  }
  return dx*dx+dy*dy>4;
}

static int _solutionStep2D(cell_2d_t *cell,struct _solution *solution)
{
  int i;
  int found=FALSE;
  
  if (cell==solution->exit) {
    double xsize= solution->para->xcellsize;
    double ysize= solution->para->ycellsize;
    double xoff=solution->para->xoff;
    double yoff=solution->para->yoff;
    appendStroke2DCell(solution->stroke,STROKE_MOVE,
		       CELL_X(cell)*xsize+xoff,CELL_Y(cell)*ysize+yoff,cell);
    found=TRUE;
  } else {
    for(i=0;i<CELL_N_NEIGHBOURS(cell);i++) {
      cell_2d_t *neighbour=CELL_NEIGHBOUR(cell,i);
      if (isOpen2D(cell,i) && visitCell(neighbour)) {
	if (_solutionStep2D(neighbour,solution)) {
	  if (_seperatedCells(neighbour,cell)) {
	    printf("seperated:%d %d\n",CELL_NO(neighbour),CELL_NO(cell));
	  }
	  double xsize= solution->para->xcellsize;
	  double ysize= solution->para->ycellsize;
	  double xoff=solution->para->xoff;
	  double yoff=solution->para->yoff;
	  appendStroke2DCell(solution->stroke,STROKE_LINE,
			     CELL_X(cell)*xsize+xoff,CELL_Y(cell)*ysize+yoff,cell);
	  found=TRUE;
	  break;
	}
      }
    }
  }
  return found;
}

stroke_t *renderMaze2DSolution(maze_2d_t *maze,render_parameter_t *para) 
{
  stroke_t *stroke=newStroke2D(MAZE_N_CELLS(maze));
  cell_2d_t *cell=MAZE_EXIT(maze);
  struct _solution solution={stroke,MAZE_ENTRANCE(maze),para};
  
  resetMaze2DUserData(maze);
  _solutionStep2D(cell,&solution);
  return solution.stroke;
}

/**********************************************************************/

stroke_t *appendStroke2DPoint(stroke_t *stroke,stroke_type_t type,double x,double y)
{
  int idx=STROKE_N_POINTS(stroke);
  if (STROKE_N_POINTS(stroke)==STROKE_MAX_POINTS(stroke)-1) {
    stroke=resizeStroke2D(stroke,STROKE_MAX_POINTS(stroke)*2);
  }
  STROKE_TYPE(stroke,idx)=type;
  STROKE_X(stroke,idx)=x;
  STROKE_Y(stroke,idx)=y;
  STROKE_CELL(stroke,idx)=NULL;
  STROKE_N_POINTS(stroke)=idx+1;
  return stroke;
}


stroke_t *appendStroke2DCell(stroke_t *stroke,stroke_type_t type,double x,double y,cell_2d_t *cell)
{
  int idx=STROKE_N_POINTS(stroke);
  if (STROKE_N_POINTS(stroke)==STROKE_MAX_POINTS(stroke)-1) {
    stroke=resizeStroke2D(stroke,STROKE_MAX_POINTS(stroke)*2);
  }
  STROKE_TYPE(stroke,idx)=type;
  STROKE_X(stroke,idx)=x;
  STROKE_Y(stroke,idx)=y;
  STROKE_CELL(stroke,idx)=cell;
  STROKE_N_POINTS(stroke)=idx+1;
  return stroke;
}

stroke_t *removeStroke2DLastPoint(stroke_t *stroke)
{
  STROKE_N_POINTS(stroke)-=1;
  return stroke;
}


stroke_t *newStroke2D(int initsize)
{
  stroke_t *retVal=VALLOC(stroke_t,initsize,single_stroke_t);
  STROKE_MAX_POINTS(retVal)=initsize;
  STROKE_N_POINTS(retVal)=0;
  return retVal; 
}

stroke_t *resizeStroke2D(stroke_t *stroke,int newsize)
{
  stroke_t *retVal=stroke;

  if (newsize>STROKE_MAX_POINTS(stroke)) {
    retVal=VREALLOC(stroke,stroke_t,newsize,single_stroke_t);
    STROKE_MAX_POINTS(retVal)=newsize;
  }
  return retVal;
}

void freeStroke2D(stroke_t *stroke)
{
  free(stroke);
}

/**********************************************************************/

void renderTextAlongStroke(maze_2d_t *maze,stroke_t *stroke,char *text)
{
  double delta,sum=0.;
  int i,idx,lidx=0;
  cell_2d_t *cur;
  int len=strlen(text);
  static char fillchars[]={"ABCEFGHIJKLMNOPQRSTUVWXYZ"};

  int n=STROKE_N_POINTS(stroke);
  if (len>n) {
    ERROR("Text too long (%d); cannot render to stroke (%d steps)",len,n);
    return;
  }
  if (n<4) {
    ERROR("Path too short (%d); stroke not rendered",n);
    return;
  }
  
  resetMaze2DUserData(maze);
  delta=(double)len/((double)n-2);
  for(i=0;i<STROKE_N_POINTS(stroke);i++) {
    cur=STROKE_CELL(stroke,i);
    if (lidx!=(int)sum) {
      CELL_USER_DATA(cur)=toupper(text[lidx]);
    } else {
      CELL_USER_DATA(cur)=-1;
    }
    lidx=(int)sum;
    sum+=delta;
  }
  for(i=0;i<MAZE_N_CELLS(maze);i++) {
    cur=MAZE_CELL(maze,i);
    if (CELL_TYPE(cur)!=TYPE_OFF && CELL_USER_DATA(cur)==0) {
      idx=rand()%n;
      if (idx<len) {
	if (rand()%3==0) {
	  CELL_USER_DATA(cur)=toupper(fillchars[rand()%strlen(fillchars)]);      
	} else {
	  CELL_USER_DATA(cur)=toupper(text[rand()%len]);      
	}
      } 
    }
  }
}

/**********************************************************************/
