/**
 *      
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2006 by Bernd Breitenbach.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 */

/**
 * \file labymain.c
 * \author Bernd Breitenbach
 * \remarks
 *      bb - Aug 14, 2006: Created.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <libgen.h>
#include <string.h>

#include "mazeintern.h"

#include <cairo.h>

enum options_e
{
  WINDINGNESS     = 256,
  DRAW_NUMBERS    = 257,
  FUZZY           = 258,
  NO_FUZZY        = 259,
  AMPLITUDE       = 260,
  TEXT            = 261,
  NO_MARKS        = 262,
  COPYRIGHT       = 263,
};


void usage(char *prog) 
{
  static char *help=""
    "usage: %s <options>\n"
    "       \n"    
    "       -t,--type=<type>         generate a maze of the given type\n"
    "           the following values are valid:\n"
    "           '3','triangular'        - a triangle based maze\n"
    "           '4','orthogonal'        - a orthogonal maze\n"
    "           '6','hexagonal','sigma' - a hexagonal maze\n"
    "           '3-4','triangle-square' - a maze from triangles & squares\n"
    "           '6-4','hexagon-square'  - a maze from hexagons & squares\n"
    "           '8-4','octagon-square'  - a maze from octagons & squares\n"
    "           'rhombic','zeta'        - a rhombus based maze\n"
    "           'cairo'                 - a cairo-tiling based maze\n"
    "           'hexagonal-rhombic'     - a maze based on rhombic divided hexagons\n"
    "           \n"
    "       -c, --columns=<number>   generate the maze with the specified\n"
    "                                no of columns\n"
    "       -r, --rows=<number>      generate the maze with the specified\n" 
    "                                no of rows\n"
    "       -o, --output=<file>      output file name; use - for stdout\n"
    "       -f, --format=<format>    the output format; if not specified\n"
    "           the extension of the file specified by -o is used\n"
    "           the following values are valid:\n"
#ifdef WITH_CAIRO_PDF
    "           'pdf'        - pdf format\n"
#endif
#ifdef WITH_CAIRO_SVG
    "           'svg'        - svg vector format\n"
#endif
#ifdef WITH_CAIRO_IMAGE
    "           'png'        - png bitmap format\n"
#endif
    "\n"
    "       -a, --algorithm=<name>   specifies the algorithm to be used\n"
    "           the following values are valid:\n"
    "           'backtracker' (default) - a recursive backtracker\n"
    "           'prim'  - prim's algorithm\n"
    "\n"
    "       -F, --foreground=<r,g,b(,a)> \n"
    "                                specify the maze color the with three (four) float values between 0. and 1.\n"
    "                                the first number specify the red component, the second the green, the  \n"
    "                                third the blue component and the optional last one the alpha value (transparency) \n"
    "       -B, --background=<r,g,b(,a)> \n"
    "                                specify the background color the with three (four) float values between 0. and 1.\n"
    "                                the first number specify the red component, the second the green, the  \n"
    "                                third the blue component and the optional last one the alpha value (transparency) \n"
    "       -e, --entrance=<dir>     where to place the entrance 'north','west','south','east'\n"
    "       -x, --exit=<dir>         where to place the exit 'north','west','south','east'\n"
#if 0
    "       -g, --goal=x,y           exclusive to exit; use cell at or near coords as goal\n"       
#endif
    "       -S, --seed=<number>      set the seed from the rng\n" 
    "       -T, --tileable           create a tileable maze\n" 
    "       -C, --cellsize=<float>   set the cell size\n" 
    "       -b, --bordersize=<float> set the border size\n" 
    "       -l, --linewidth=<float>  set the line width\n" 
    "       --nomarks                omit drawing of entrance/exit marks\n" 
    "       -P, --solution           draw the solution path\n" 
    "       --numbers                draw cell numbers (for debugging purpose)\n" 
    "       --text=<text>            draw the text specified along the solution\n" 
    "                                and fill random characters into the maze\n"
    "       --fuzzy                  draw fuzzy strokes\n" 
    "       --amplitude=<float>      amplitude for fuzzy strokes\n" 
    "       --no-fuzzy=<number>      number of fuzzy substrokes\n" 
    "       -m, --mask=<pngfile>     specifies mask to be used for generating the maze\n"
    "       -v, --version            print the program version\n"
    "       -d, --dummy              don't create a maze, render a grid only\n"
    "       --copyright              insert a copyright message into the maze\n"
#if 0
    "       backtrack:\n"
    "       --windingness=<[0..1]>   determines the windingness of the path \n"
    "           0.5 is defualt, values below results in a\n"
    "           more winding path\n"
#endif
    "       \n"
    "";
  fprintf(stderr,help,prog);
  exit(-1);
}


void version(char *prog) 
{
  fprintf(stderr,"%s " VERSION "\n"
	  "Copyright (C) 2006 Bernd Breitenbach\n"
	  "This is free software.  You may redistribute copies of it under the terms of\n"
	  "the GNU General Public License <http://www.gnu.org/licenses/gpl.html>.\n"
	  "There is NO WARRANTY, to the extent permitted by law.\n"
	  "\n"
	  "Written by Bernd Breitenbach.\n",prog);
}


static char *short_options="dt:TPg:c:r:o:a:e:x:S:C:b:l:m:f:F:B:v";
static struct option long_options[]=
{
  {"type"           ,1,NULL,'t'},
  {"format"         ,1,NULL,'f'},
  {"entrance"       ,1,NULL,'e'},
  {"exit"           ,1,NULL,'x'},
#if 0
  {"goal"           ,1,NULL,'g'},
#endif
  {"tileable"       ,0,NULL,'T'},
  {"columns"        ,1,NULL,'c'},
  {"rows"           ,1,NULL,'r'},
  {"output"         ,1,NULL,'o'},
  {"seed"           ,1,NULL,'S'},
  {"cellsize"       ,1,NULL,'C'},
  {"bordersize"     ,1,NULL,'b'},
  {"background"     ,1,NULL,'B'},
  {"foreground"     ,1,NULL,'F'},
  {"numbers"        ,0,NULL,DRAW_NUMBERS},
  {"nomarks"        ,0,NULL,NO_MARKS},
  {"solution"       ,0,NULL,'P'},
  {"linewidth"      ,1,NULL,'l'},
  {"algorithm"      ,1,NULL,'a'},
  {"mask"           ,1,NULL,'m'},
  {"windingness"    ,1,NULL,WINDINGNESS},
  {"fuzzy"          ,0,NULL,FUZZY},
  {"no-fuzzy"       ,1,NULL,NO_FUZZY},
  {"amplitude"      ,1,NULL,AMPLITUDE},
  {"text"           ,1,NULL,TEXT},
  {"version"        ,0,NULL,'v'},
  {"dummy"          ,0,NULL,'d'},
  {"copyright"      ,0,NULL,COPYRIGHT},
  {NULL             ,0,NULL,0}
};

struct _algorithm
{
  char           *name;
  carver_2d_t    *(*carver)(struct _maze_parameter *);
} algorithms[2]={
  {"backtracker",&newRandomBacktracker2D},
  {"prim"       ,&newPrimCarver2D},
};


struct _directions
{
  char         *name;
  direction_t  dir;
} directions[4] ={
  {"north",NORTH},
  {"west" ,WEST},
  {"south",SOUTH},
  {"east" ,EAST}
};

/**********************************************************************/

static maze_parameter_t parameter = {
  .columns    =-1,
  .rows       =-1,
  .entrance   =NO_DIRECTION,
  .exit       =NO_DIRECTION,
  .goal       ={{0.,0.,0.}},
  .structure  =NULL,
  .seed       =-1,
  .windingness=0.5,
  .tileable   =FALSE,
  .construct  =NULL,
  .carver     =NULL,
  .renderer   =NULL,
  .render_idx =RENDER_NONE,
  .mask       =NULL
};

static render_parameter_t render_para={
  .render=NULL,
  .filename="",
  .width=-1.,
  .height=-1.,
  .xcellsize=16.,
  .ycellsize=16.,
  .bordersize=16.,
  .xoff=16.,
  .yoff=16.,
  .linewidth=2.5,
  .drawno=FALSE,
  {0.,0.,0.8,0.5},
  {1.,1.,1. ,1.},
  .draw_marks=TRUE,
  .draw_solution=FALSE,
  .fuzzy=FALSE,
  .amplitude=0.1,
  .n_fuzzy=8,
  .solution_text=NULL,
  .copyright_text=NULL
};

static int convert_to_int(char *value,int *ival)
{
  char *end;
  *ival=(int)strtol(value,&end,0);
  return *end!='\0';
}

static int convert_to_double(char *value,double *dval)
{
  char *end;
  *dval=(double)strtod(value,&end);
  return *end!='\0';
}

static int convert_to_points(char *value,double *dval)
{
  int retVal=0;
  char *end;
  *dval=(double)strtod(value,&end);
  if (*end!='\0') {
    if (strcmp(end,"px")==0) {
      /* nothing! */
    } else if (strcmp(end,"pt")==0) {
      /* nothing! */
    } else if (strcmp(end,"mm")==0) {
      *dval/=0.35277138;
    } else if (strcmp(end,"cm")==0) {
      *dval/=0.035277138;
    } else if (strcmp(end,"in")==0) {
      *dval*=72;
    } else {
      retVal=1;
    }
  }
  return retVal;
}

static int convert_to_color(char *value,color_t *cval)
{
  int retVal=0;
  char *end;
  cval->red=(double)strtod(value,&end);
  retVal=(*end!=',');
  if(!retVal) {
    cval->green=(double)strtod(end+1,&end);
    retVal=(*end!=',');
  }
  if(!retVal) {
    cval->blue=(double)strtod(end+1,&end);
    if (*end==',') {
      cval->alpha=(double)strtod(end+1,&end);
    } else {
      cval->alpha=1.;    
    }
  }
  return retVal;
}

static int convert_to_coords(char *value,coords_u *coords)
{
  
  char *end;
  coords->cartesic.x=(double)strtod(value,&end);
  if (*end==',') {
    coords->cartesic.y=(double)strtod(end,&end);
  }
  if (*end==',') {
    coords->cartesic.z=(double)strtod(end,&end);
  }
  return *end!='\0';
}



static int convert_to_direction(char *value,int *ival)
{
  int retVal=1;
  int i;
  for(i=0;i<sizeof(directions)/sizeof(struct _directions);i++) {
    if (strcmp(value,directions[i].name)==0) {
      *ival=directions[i].dir;
      retVal=0;
      break;
    }
  }
  return retVal;
}


static laby_render_f renderFuncs[]={
  NULL, 
#ifdef WITH_CAIRO_PDF
  &renderPDF2D,
#else
  NULL,
#endif
#ifdef WITH_CAIRO_PDF
  &renderPNG2D,
#else
  NULL,
#endif
#ifdef WITH_CAIRO_PDF
  &renderSVG2D,
#else
  NULL,
#endif
};

/**********************************************************************/
int main(int argc,char **argv) 
{
  maze_2d_t *laby=NULL;
  carver_2d_t *carver=NULL;
  int i;
  char *ptr;
  int doexit=0;
  int dummy=FALSE;

  parameter.construct=&new2DMaze;
  parameter.carver=NULL; 
  parameter.renderer=renderFuncs;

  int c;
  while((c=getopt_long(argc,argv,short_options,long_options,NULL))>0) {
    switch(c) {
    case 't':
      if (strcmp(optarg,"3")==0 ||strcmp(optarg,"triangular")==0) {
	parameter.structure=&tria_struct;
      } else if (strcmp(optarg,"4")==0 ||strcmp(optarg,"orthogonal")==0) {
	parameter.structure=&ortho_struct;
      } else if (strcmp(optarg,"6")==0 ||strcmp(optarg,"hexagonal")==0 ||
		 strcmp(optarg,"sigma")==0) {
	parameter.structure=&hexa_struct;
      } else if (strcmp(optarg,"6-4")==0 ||strcmp(optarg,"hexagon-square")==0) {
	parameter.structure=&hexa_square_struct;
      } else if (strcmp(optarg,"3-4")==0 ||strcmp(optarg,"triangle-square")==0) {
	parameter.structure=&triangle_square_struct;
      } else if (strcmp(optarg,"8-4")==0 ||strcmp(optarg,"octagon-square")==0) {
	parameter.structure=&octa_square_struct;
      } else if (strcmp(optarg,"hexagonal-rhombic")==0) {
	parameter.structure=&hexa_rhombic_struct;
      } else if (strcmp(optarg,"rhombic")==0 || strcmp(optarg,"zeta")==0) {
	parameter.structure=&rhombic_struct;
      } else if (strcmp(optarg,"cairo")==0) {
	parameter.structure=&cairo_struct;
      } else {
	fprintf(stderr,"unknown maze type '%s'\n",optarg);
	doexit=1;
      }
      break;
    case 'T':
      parameter.tileable=TRUE;
      break;
    case 'P':
      render_para.draw_solution=TRUE;
      break;
    case 'c':
      if (convert_to_int(optarg,&parameter.columns)!=0) {
	fprintf(stderr,"invalid column parameter '%s'\n",optarg);
	doexit=1;
      }
      break;
    case 'r':
      if (convert_to_int(optarg,&parameter.rows)!=0) {
	fprintf(stderr,"invalid row parameter '%s'\n",optarg);
	doexit=1;
      }
      break;
    case 'e':
      if (parameter.entrance!=NO_DIRECTION) {
	fprintf(stderr,"illegal combination of entrance/exit/goal option\n");
	doexit=1;
      } else if (convert_to_direction(optarg,&parameter.entrance)!=0) {
	fprintf(stderr,"invalid direction '%s'\n",optarg);
	doexit=1;
      }
      break;
    case 'x':
      if (parameter.exit!=NO_DIRECTION) {
	fprintf(stderr,"illegal combination of entrance/exit/goal option\n");
	doexit=1;
      } else if (convert_to_direction(optarg,&parameter.exit)!=0) {
	fprintf(stderr,"invalid direction '%s'\n",optarg);
	doexit=1;
      }
      break;
    case 'g':
      if (parameter.exit!=NO_DIRECTION) {
	fprintf(stderr,"illegal combination of entrance/exit/goal option\n");
	doexit=1;
      } else if (convert_to_coords(optarg,&parameter.goal)!=0) {
	fprintf(stderr,"invalid coords '%s'\n",optarg);
	doexit=1;
      }      
      break;
    case 'f':
      if (0) {
#ifdef WITH_CAIRO_IMAGE
      } else if (strcmp(optarg,"png")==0) {
	parameter.render_idx=RENDER_PNG;
#endif
#ifdef WITH_CAIRO_SVG
      } else if (strcmp(optarg,"svg")==0) {
	parameter.render_idx=RENDER_SVG;
#endif
#ifdef WITH_CAIRO_PDF
      } else if (strcmp(optarg,"pdf")==0) {
	parameter.render_idx=RENDER_PDF;
#endif
      } else {
	fprintf(stderr,"unknown output format '%s'\n",optarg);
	doexit=1;
      }
      break;
    case 'o':
      render_para.filename=strdup(optarg);
      if (strcmp(optarg,"-")==0) {
	render_para.filename=NULL;
      } else {
	ptr=optarg+strlen(optarg)-4;
	if (parameter.render_idx==RENDER_NONE) {
	  if (0) {
#ifdef WITH_CAIRO_IMAGE
	  } else if (strcasestr(optarg,".png")==ptr) {
	    parameter.render_idx=RENDER_PNG;
#endif
#ifdef WITH_CAIRO_SVG
	  } else if (strcasestr(optarg,".svg")==ptr) {
	    parameter.render_idx=RENDER_SVG;
#endif
#ifdef WITH_CAIRO_PDF
	  } else if (strcasestr(optarg,".pdf")==ptr) {
	    parameter.render_idx=RENDER_PDF;
#endif
	  }
	}
      }
      break;
    case NO_MARKS:
      render_para.draw_marks=FALSE;
      break;
    case DRAW_NUMBERS:
      render_para.drawno=TRUE;
      break;
    case 'a':
      for(i=0;i<sizeof(algorithms)/sizeof(struct _algorithm);i++) {
	if (strcmp(optarg,algorithms[i].name)==0) {
	  parameter.carver=algorithms[i].carver;
	}
      }
      if (parameter.carver==NULL) {
	fprintf(stderr,"unknown algorithm '%s'\n",optarg);
	doexit=1;
      }
      break;
    case 'm':
      parameter.mask=newMaze2DMask(optarg);
      if (parameter.mask==NULL) {
	fprintf(stderr,"cannot load mask file '%s'\n",optarg);
	doexit=TRUE;
      }
      break;
    case 'S':
      if (convert_to_int(optarg,&parameter.seed)!=0) {
	fprintf(stderr,"invalid seed parameter '%s'\n",optarg);
	doexit=1;
      }
      break;
    case 'C':
      if (convert_to_points(optarg,&render_para.xcellsize)!=0) {
	fprintf(stderr,"invalid cellsize parameter '%s'\n",optarg);
	doexit=1;
      }
      render_para.ycellsize=render_para.xcellsize;
      break;
    case 'b':
      if (convert_to_points(optarg,&render_para.bordersize)!=0) {
	fprintf(stderr,"invalid bordersize parameter '%s'\n",optarg);
	doexit=1;
      }
      break;
    case 'B':
      if (convert_to_color(optarg,&render_para.background)!=0) {
	fprintf(stderr,"invalid color '%s'\n",optarg);
	doexit=1;
      }
      break;
    case 'F':
      if (convert_to_color(optarg,&render_para.foreground)!=0) {
	fprintf(stderr,"invalid color '%s'\n",optarg);
	doexit=1;
      }
      break;
    case 'l':
      if (convert_to_points(optarg,&render_para.linewidth)!=0) {
	fprintf(stderr,"invalid linewidth parameter '%s'\n",optarg);
	doexit=1;
      }
      break;
    case WINDINGNESS:
      if (convert_to_double(optarg,&parameter.windingness)!=0) {
	fprintf(stderr,"invalid cellsize parameter '%s'\n",optarg);
	doexit=1;
      }
      break;
    case AMPLITUDE:
      if (convert_to_double(optarg,&render_para.amplitude)!=0) {
	fprintf(stderr,"invalid amplitude parameter '%s'\n",optarg);
	doexit=1;
      }
      render_para.fuzzy=TRUE;
      break;
    case NO_FUZZY:
      if (convert_to_int(optarg,&render_para.n_fuzzy)!=0) {
	fprintf(stderr,"invalid fuzzy parameter '%s'\n",optarg);
	doexit=1;
      }
      render_para.fuzzy=TRUE;
      break;
    case COPYRIGHT:
      render_para.copyright_text="Maze 5 (c) B. Breitenbach";
      break;
    case TEXT:
      render_para.solution_text=strdup(optarg);
      break;
    case FUZZY:
      render_para.fuzzy=TRUE;
      break;
    case 'v':
      version(basename(argv[0]));
      exit(0);
      break;
    case 'd':
      dummy=TRUE;
      break;
    case '?':
      doexit=1;
      break;
    }
  }

  if (argv[optind]!=NULL) {
    fprintf(stderr,"illegal argument '%s'\n",argv[optind]);
    doexit=1;
  }
  if (parameter.render_idx!=RENDER_NONE) {
    if (parameter.renderer==NULL) {
      fprintf(stderr,"no labyrinth type specified\n");
      doexit=1;
    }
  } else {
    fprintf(stderr,"no output format specified\n");
    doexit=1;
  }
  if (doexit){
    usage(basename(argv[0]));
  }
  
  if (parameter.structure==NULL) {
      parameter.structure=&ortho_struct;
  }
  if (parameter.carver==NULL) {
    parameter.carver=&newRandomBacktracker2D;
  }
  laby=parameter.construct(&parameter);
  if (!dummy) {
    carver=parameter.carver(&parameter);
    carveMaze2D(carver,laby);
  }
  parameter.renderer[parameter.render_idx](laby,&render_para);
  return 0;
}
