/*
 * @(#) $Id: turing-pattern.c 340 2006-01-03 00:20:39Z yeti $
 * This is a plugin for the GIMP.
 *
 * Copyright (C) 2004 Yeti (David Necas) <yeti@physics.muni.cz>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/*
 * This plug-in renderes `papillary-lines' Turing pattern and a few morw
 * queer textures.
 *
 * Please send patches, suggestions, and other invectives to me:
 * <yeti@physics.muni.cz>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#else
/* To include libgimp/stdplugins-intl.h */
#  define GETTEXT_PACKAGE "gimp20"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include "libgimp/stdplugins-intl.h"

#define PROCEDURE_NAME   "turing_pattern"
#define DATA_KEY_VALS    "plug_in_turing_pattern"

enum {
    BITS_PER_SAMPLE   = 8,
    CHANNEL_MAX_VALUE = (1U << BITS_PER_SAMPLE) - 1,

    PREVIEW_SIZE = 128,
    SCALE_WIDTH = 201,
    CELL_SIZE_WIDTH = 84,

    /* FIXME: Is it still impossible to store true unlimited strings as
     * parameters? */
    GRADIENT_NAME_SIZE = 256,
    GRADIENT_SAMPLE_SIZE = 8192,

    /* How often progressbar is updated */
    PROGRESS_STEP = 20,
    PROGRESS_PREVIEW_STEP = 100
};

/* Empiric factor translating integration parameter h to feature size */
#define FEATURE_SIZE_TO_H     0.0462

/* Funny nonlinear function.  It has following properties
 * - odd (symmetry about origin)
 * - large positive derivation in 0
 * - a maxima at some positive value
 * - zero at some larger value
 * - then negative, but not too much
 */
#define F(x) ((x)/(1 + 0.01*(x)*(x)) - 0.01*(x))

enum {
    RESPONSE_REVERT = 1,
    RESPONSE_RESET
};

enum {
    RECOMPUTE_ALL   = 1 << 0,
    RECOMPUTE_COLOR = 1 << 1
};

typedef enum {
    COLORING_METHOD_GRAY     = 0,
    COLORING_METHOD_GRADIENT = 1
} ColoringMethod;

typedef enum {
    PATTERN_PAPILLARY,
    PATTERN_CONGLOMERATE,
    PATTERN_PROTOPLASM,
    PATTERN_CHEETAH
} PatternType;

typedef struct {
    const gchar *name;
    guint        value;
} Enum;

typedef struct {
    gfloat min;
    gfloat max;
} DataRange;

typedef struct {
    gint xres;
    gint yres;
    gfloat h;
    gint iter;
    gint iscale;
} Stage;

typedef struct {
    gdouble        feature_size;
    ColoringMethod cm;
    PatternType    type;
    gchar          grad[GRADIENT_NAME_SIZE];
} PluginValues;

typedef struct {
    GtkObject *feature_size;
    GtkWidget *type;
    GtkWidget *cm_gray;
    GtkWidget *cm_grad;
    GtkWidget *grad;
    GtkWidget *update;
} PluginControls;

static void      query                  (void);
static void      run                    (const gchar *name,
                                         gint nparams,
                                         const GimpParam *param,
                                         gint *nreturn_vals,
                                         GimpParam **return_vals);
static gboolean  plugin_dialog          (void);
static void      gradient_changed       (const gchar *name,
                                         gint width,
                                         const gdouble *grad_data,
                                         gboolean dialog_closing,
                                         gpointer user_data);
static void      coloring_method_changed(GtkWidget *widget,
                                         gpointer data);
static void      refresh_controls       (const PluginValues *new_pvals);
static void      commit                 (GimpDrawable *drawable,
                                         gint channel);
static void      commit_one_tile        (GimpPixelRgn *rgn,
                                         gint channel,
                                         gint gradwidth);
static void      add_transparency       (gint channel);
static void      render_gradient        (void);
static void      pattern_type_changed   (GtkWidget *combo,
                                         PatternType *valp);
static void      execute_all            (GimpDrawable *drawable);
static void      execute_color          (GimpDrawable *drawable);
static void      execute                (GimpDrawable *drawable,
                                         guint recompute);
static void      turingp_tp             (void);
static void      iterate_tp             (gint width,
                                         gint height,
                                         gint iterations,
                                         gfloat h,
                                         gint iscale);
static void      do_iter_tp             (gint xres,
                                         gint yres,
                                         gfloat *data,
                                         gfloat h);
static void      turingp_cr             (void);
static void      iterate_cr             (gint width,
                                         gint height,
                                         gint iterations);
static void      do_iter_cr             (gint xres,
                                         gint yres,
                                         gfloat *data);
static void      turingp_pc             (gint channel);
static void      iterate_pc             (gint width,
                                         gint height,
                                         gfloat p,
                                         gfloat q,
                                         gfloat p0,
                                         gfloat q0,
                                         gint iterations);
static void      do_iter_pc             (gint xres,
                                         gint yres,
                                         gfloat *data,
                                         gfloat p,
                                         gfloat q,
                                         gfloat p0,
                                         gfloat q0);
static void      scale_up_periodic      (gint sxres,
                                         gint syres,
                                         gfloat *src,
                                         gint dxres,
                                         gint dyres,
                                         gfloat *dst);
static void      checker_smooth         (gint xres,
                                         gint yres,
                                         gfloat *r0,
                                         gfloat *r1,
                                         gfloat *prr0,
                                         gfloat *prr1);
static DataRange normalize              (gint n,
                                         gfloat *data);
static DataRange normalize_hist         (gint n,
                                         gfloat *data);

static GimpPlugInInfo PLUG_IN_INFO = {
    NULL,  /* init_proc  */
    NULL,  /* quit_proc  */
    query, /* query_proc */
    run,   /* run_proc   */
};

static const PluginValues pvals_defaults = {
    10.0,                 /* feature_size */
    COLORING_METHOD_GRAY, /* mode         */
    PATTERN_PAPILLARY,    /* type         */
    "",                   /* grad         */
};

static PluginControls pctrl;
static PluginValues pvals; /* = pvals_defaults; */
static PluginValues pvals_old; /* for revert */

static GRand *rng = NULL;
static GtkWidget *preview_image = NULL;
static GdkPixbuf *preview = NULL;
static GtkWidget *preview_progressbar = NULL;
static gboolean preview_mode;
static gboolean is_rgb;
static GimpDrawable *drawable;
static gdouble *gradient = NULL;
static gint gradient_p_width = -1;

static gint ix1, iy1, ix2, iy2;     /* Selected image size. */
static gint bpp, has_alpha, alpha;
static glong progress;
static gdouble max_progress;
static DataRange data_range;
static gfloat *pdata = NULL;


MAIN()

static void
query(void)
{
    static GimpParamDef args[] = {
        { GIMP_PDB_INT32,    "run_mode", "Interactive, non-interactive" },
        { GIMP_PDB_IMAGE,    "image", "Input image (unused)" },
        { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },
        { GIMP_PDB_INT32,    "type", "Pattern type: { PAPILLARY (0), CONGLOMERATE (1), CHEETAH (2), PROTOPLASM (3) }" },
        { GIMP_PDB_INT32,    "cm", "Coloring mode: { GRAYSCALE (0), GRADIENT (1) }" },
        { GIMP_PDB_FLOAT,    "feature_size", "Characteristic feature size" },
        { GIMP_PDB_STRING,   "grad", "Gradient used for coloring (when cm is GRADIENT)" },
    };

    gimp_install_procedure
        (DATA_KEY_VALS,
         "Renderers kinds of Turing pattern",
         "This plug-in renderes a one type of Turing pattern (papillary "
         "lines) and a few not-so-Turing patterns.",
         "Yeti",
         "Yeti",
         "Nov 2004",
         N_("<Image>/Filters/Render/Pattern/_Turing Pattern..."),
         "RGB*, GRAY*",
         GIMP_PLUGIN,
         G_N_ELEMENTS(args), 0, args, NULL);
}

static void
run(const gchar *name,
    gint nparams,
    const GimpParam  *param,
    gint *nreturn_vals,
    GimpParam  **return_vals)
{
    static GimpParam values[1];
    GimpRunMode run_mode;
    GimpPDBStatusType status = GIMP_PDB_SUCCESS;

    run_mode = param[0].data.d_int32;

    bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
    bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
#endif
    textdomain(GETTEXT_PACKAGE);

    *nreturn_vals = 1;
    *return_vals = values;

    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = status;

    drawable = gimp_drawable_get(param[2].data.d_drawable);
    is_rgb = gimp_drawable_is_rgb(drawable->drawable_id);
    has_alpha = gimp_drawable_has_alpha(drawable->drawable_id);

    pvals = pvals_defaults;
    rng = g_rand_new();
    switch (run_mode) {
        case GIMP_RUN_INTERACTIVE:
        gimp_get_data(DATA_KEY_VALS, &pvals);

        pvals_old = pvals;
        if (!plugin_dialog()) {
            gimp_drawable_detach(drawable);
            return;
        }
        break;

        case GIMP_RUN_NONINTERACTIVE:
        if (nparams != 7)
            status = GIMP_PDB_CALLING_ERROR;
        else {
            pvals.type = (gint)param[3].data.d_int32;
            pvals.cm = (gint)param[4].data.d_int32;
            pvals.feature_size = (gint)param[5].data.d_float;
            g_strlcpy(pvals.grad, (gchar*)param[6].data.d_string,
                      sizeof(pvals.grad));
        }
        break;

        case GIMP_RUN_WITH_LAST_VALS:
        gimp_get_data(DATA_KEY_VALS, &pvals);
        break;

        default:
        break;
    }

    if (status == GIMP_PDB_SUCCESS) {
        if (is_rgb || gimp_drawable_is_gray(drawable->drawable_id)) {
            gimp_progress_init(_("Turing Pattern..."));

            preview_mode = FALSE;
            execute_all(drawable);

            if (run_mode != GIMP_RUN_NONINTERACTIVE)
                gimp_displays_flush();

            if (run_mode == GIMP_RUN_INTERACTIVE
                || run_mode == GIMP_RUN_WITH_LAST_VALS) {
                g_strlcpy(pvals.grad, gimp_context_get_gradient(),
                          sizeof(pvals.grad));
                gimp_set_data(DATA_KEY_VALS, &pvals, sizeof(PluginValues));
            }
        }
        else
            status = GIMP_PDB_EXECUTION_ERROR;
    }

    values[0].data.d_status = status;
    gimp_drawable_detach(drawable);
}

/* remove some source code redundancies {{{ */
static inline void
yeti_table_attach_single_cell(GtkWidget *table,
                              GtkWidget *child,
                              gint column,
                              gint row)
{
    gtk_table_attach(GTK_TABLE(table), child, column, column+1, row, row+1,
                     GTK_EXPAND|GTK_FILL, 0, 0, 0);
}


static inline GtkWidget*
yeti_preview_new_with_progress(GtkWidget *hbox,
                               GimpDrawable *drawable,
                               GCallback callback,
                               GtkWidget **progressbar,
                               GtkWidget **button)
{
    GtkWidget *frame, *vbox, *image;
    GdkPixbuf *pixbuf;

    vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
    gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);

    pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB,
                            FALSE,
                            BITS_PER_SAMPLE,
                            PREVIEW_SIZE,
                            PREVIEW_SIZE);
    gdk_pixbuf_fill(pixbuf, 0);

    image = gtk_image_new_from_pixbuf(pixbuf);
    g_object_unref(pixbuf);
    gtk_container_add(GTK_CONTAINER(frame), image);

    *progressbar = gtk_progress_bar_new();
    gtk_widget_set_size_request(*progressbar, PREVIEW_SIZE, 8);
    gtk_box_pack_start(GTK_BOX(vbox), *progressbar, FALSE, FALSE, 0);

    *button = gtk_button_new_with_mnemonic(_("_Update Preview"));
    gtk_box_pack_start(GTK_BOX(vbox), *button, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(*button), "clicked",
                     G_CALLBACK(callback), drawable);

    return image;
}

static inline GtkObject*
yeti_scale_entry_new_double(const gchar *name,
                            GtkWidget *table,
                            GtkSizeGroup *sgroup,
                            gint row,
                            gdouble *value,
                            gdouble min,
                            gdouble max,
                            gdouble step,
                            gdouble page,
                            gint digits)
{
    GtkObject *adj;

    *value = CLAMP(*value, min, max);
    adj = gimp_scale_entry_new(GTK_TABLE(table), 0, row, name, SCALE_WIDTH, 0,
                               *value, min, max, step, page, digits,
                               TRUE, 0, 0, NULL, NULL);
    g_signal_connect(adj, "value-changed",
                     G_CALLBACK(gimp_double_adjustment_update), value);
    if (sgroup)
        gtk_size_group_add_widget(sgroup, GIMP_SCALE_ENTRY_SPINBUTTON(adj));

    return adj;
}

static inline void
yeti_coloring_method_attach(GtkWidget *table,
                            gint *row,
                            const gchar *gradlabel,
                            GtkWidget **grad,
                            GtkWidget **cm_gray,
                            GtkWidget **cm_grad,
                            ColoringMethod *cm)
{
    GtkRadioButton *group;
    GtkWidget *align;

    if (!is_rgb)
        *cm = COLORING_METHOD_GRAY;

    *grad = gimp_gradient_select_widget_new(gradlabel,
                                            gimp_context_get_gradient(),
                                            gradient_changed,
                                            NULL);
    align = gtk_alignment_new(0.0, 0.5, 0.0, 0.0);
    gtk_container_add(GTK_CONTAINER(align), *grad);
    yeti_table_attach_single_cell(table, align, 1, *row+1);
    gtk_widget_set_sensitive(*grad, *cm == COLORING_METHOD_GRADIENT);

    *cm_gray = gtk_radio_button_new_with_mnemonic(NULL, _("Gra_yscale"));
    group = GTK_RADIO_BUTTON(*cm_gray);
    g_signal_connect(*cm_gray, "toggled",
                     G_CALLBACK(coloring_method_changed),
                     GUINT_TO_POINTER(COLORING_METHOD_GRAY));
    yeti_table_attach_single_cell(table, *cm_gray, 0, *row);

    *cm_grad = gtk_radio_button_new_with_mnemonic_from_widget(group,
                                                              _("_Gradient"));
    if (*cm == COLORING_METHOD_GRADIENT)
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(*cm_grad), TRUE);
    g_signal_connect(*cm_grad, "toggled",
                     G_CALLBACK(coloring_method_changed),
                     GUINT_TO_POINTER(COLORING_METHOD_GRADIENT));
    yeti_table_attach_single_cell(table, *cm_grad, 0, *row+1);
    if (!is_rgb)
        gtk_widget_set_sensitive(*cm_grad, FALSE);

    *row += 2;
}
/* }}} */

static gboolean
plugin_dialog(void)
{
    GtkWidget *dlg, *hbox, *table, *label, *align;
    gint run, row;
    gboolean ok;

    preview_mode = TRUE;
    gimp_ui_init(PROCEDURE_NAME, TRUE);

    dlg = gimp_dialog_new(_("Turing Pattern"), PROCEDURE_NAME,
                          NULL, 0,
                          gimp_standard_help_func,
                          "filters/turing_pattern.html",

                          _("Re_vert"), RESPONSE_REVERT,
                          GIMP_STOCK_RESET, RESPONSE_RESET,
                          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                          GTK_STOCK_OK, GTK_RESPONSE_OK,
                          NULL);

    hbox = gtk_hbox_new(FALSE, 8);
    gtk_container_set_border_width(GTK_CONTAINER(hbox), 6);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), hbox,
                       TRUE, TRUE, 0);

    preview_image = yeti_preview_new_with_progress(hbox, drawable,
                                                   G_CALLBACK(execute_all),
                                                   &preview_progressbar,
                                                   &pctrl.update);
    preview = gtk_image_get_pixbuf(GTK_IMAGE(preview_image));

    table = gtk_table_new(3, 4, FALSE);
    gtk_table_set_col_spacings(GTK_TABLE(table), 4);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    row = 0;

    align = gtk_alignment_new(0.0, 0.0, 1.0, 0.0);
    gtk_container_add(GTK_CONTAINER(align), table);
    gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 0);

    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    row = 0;

    yeti_coloring_method_attach(table, &row, _("Turing Pattern Gradient"),
                                &pctrl.grad, &pctrl.cm_gray, &pctrl.cm_grad,
                                &pvals.cm);

    pctrl.type
        = gimp_int_combo_box_new(_("Papillary Lines"),   PATTERN_PAPILLARY,
                                 _("Conglomerate Rock"), PATTERN_CONGLOMERATE,
                                 _("Protoplasm"),        PATTERN_PROTOPLASM,
                                 _("Cheetah"),           PATTERN_CHEETAH,
                                 NULL);
    gtk_combo_box_set_wrap_width(GTK_COMBO_BOX(pctrl.type), 1);
    gtk_table_attach(GTK_TABLE(table), pctrl.type, 1, 2, row, row + 1,
                     GTK_EXPAND | GTK_FILL, 0, 2, 2);
    gimp_int_combo_box_set_active(GIMP_INT_COMBO_BOX(pctrl.type), pvals.type);
    g_signal_connect(pctrl.type, "changed",
                     G_CALLBACK(pattern_type_changed), &pvals.type);

    label = gtk_label_new_with_mnemonic(_("Pattern _type:"));
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), label, 0, 1, row, row + 1,
                     GTK_EXPAND | GTK_FILL, 0, 2, 2);
    gtk_label_set_mnemonic_widget(GTK_LABEL(label), pctrl.type);
    row++;

    pctrl.feature_size = yeti_scale_entry_new_double(_("_Feature size:"),
                                                     table, NULL, row,
                                                     &pvals.feature_size,
                                                     5.0, 25.0, 0.5, 5.0, 1);
    gimp_scale_entry_set_sensitive(pctrl.feature_size,
                                   pvals.type == PATTERN_PAPILLARY);
    row++;

    refresh_controls(&pvals);
    gtk_widget_show_all(dlg);

    ok = FALSE;
    do {
        run = gimp_dialog_run(GIMP_DIALOG(dlg));
        switch (run) {
            case RESPONSE_REVERT:
            refresh_controls(&pvals_old);
            break;

            case RESPONSE_RESET:
            refresh_controls(&pvals_defaults);
            break;

            default:
            ok = TRUE;
            break;
        }
    } while (!ok);

    gtk_widget_destroy(dlg);

    return run == GTK_RESPONSE_OK;
}

static void
pattern_type_changed(GtkWidget *combo,
                     PatternType *valp)
{
    gimp_int_combo_box_get_active(GIMP_INT_COMBO_BOX(combo), (gint*)valp);
    gimp_scale_entry_set_sensitive(pctrl.feature_size,
                                   *valp == PATTERN_PAPILLARY);
}

static void
coloring_method_changed(GtkWidget *widget,
                        gpointer data)
{
    ColoringMethod cm;

    if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)))
        return;

    cm = GPOINTER_TO_UINT(data);
    if (pvals.cm == cm)
        return;
    gtk_widget_set_sensitive(pctrl.grad, cm == COLORING_METHOD_GRADIENT);
    pvals.cm = cm;
    execute_color(drawable);
}

static void
gradient_changed(const gchar *name,
                 gint width,
                 const gdouble *grad_data,
                 gboolean dialog_closing,
                 gpointer user_data)
{
    gimp_context_set_gradient(name);
    gradient = g_realloc(gradient, width*sizeof(gdouble));
    memcpy(gradient, grad_data, width*sizeof(gdouble));
    gradient_p_width = width/4;

    execute_color(drawable);
}

static void
refresh_controls(const PluginValues *new_pvals)
{
    pvals.feature_size = new_pvals->feature_size;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.feature_size),
                             pvals.feature_size);

    pvals.type = new_pvals->type;
    gimp_int_combo_box_set_active(GIMP_INT_COMBO_BOX(pctrl.type), pvals.type);

    pvals.cm = new_pvals->cm;
    if (!is_rgb)
        pvals.cm = COLORING_METHOD_GRAY;
    switch (pvals.cm) {
        case COLORING_METHOD_GRAY:
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.cm_gray), TRUE);
        break;

        case COLORING_METHOD_GRADIENT:
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.cm_grad), TRUE);
        break;
    }
    gtk_widget_set_sensitive(pctrl.grad, pvals.cm == COLORING_METHOD_GRADIENT);
}

/****************************************************************************/

static void
execute_all(GimpDrawable *drawable)
{
    execute(drawable, RECOMPUTE_ALL);
}

static void
execute_color(GimpDrawable *drawable)
{
    execute(drawable, RECOMPUTE_COLOR);
}

static void
execute(GimpDrawable *drawable,
        guint recompute)
{
    gint i, chmax;

    if (preview_mode && recompute == RECOMPUTE_COLOR && !pdata)
        return;

    if (preview_mode)
        gtk_widget_set_sensitive(pctrl.update, FALSE);

    /* compute sizes and other stuff of this kind */
    alpha = -1000; /* to cause a segfault ;-> */
    if (preview_mode) {
        ix1 = iy1 = 0;
        ix2 = gdk_pixbuf_get_width(preview);
        iy2 = gdk_pixbuf_get_height(preview);
        bpp = gdk_pixbuf_get_n_channels(preview);
    }
    else {
        gimp_drawable_mask_bounds(drawable->drawable_id,
                                  &ix1, &iy1, &ix2, &iy2);
        bpp = gimp_drawable_bpp(drawable->drawable_id);
        if (has_alpha)
            alpha = bpp-1;
    }
    chmax = bpp;

    render_gradient();
    if (recompute == RECOMPUTE_ALL || !pdata) {
        g_free(pdata);
        pdata = NULL;

        switch (pvals.type) {
            case PATTERN_PAPILLARY:
            turingp_tp();
            break;

            case PATTERN_CONGLOMERATE:
            turingp_cr();
            break;

            case PATTERN_PROTOPLASM:
            turingp_pc(0);
            break;

            case PATTERN_CHEETAH:
            turingp_pc(1);
            break;

            default:
            g_assert_not_reached();
            break;
        }
    }

    for (i = 0; i < chmax; i++)
        commit(drawable, i);

    if (has_alpha && preview_mode && pvals.cm == COLORING_METHOD_GRADIENT) {
        for (i = 0; i < chmax; i++)
            add_transparency(i);
    }


    if (preview_mode) {
        gtk_widget_queue_draw(preview_image);
        gtk_widget_set_sensitive(pctrl.update, TRUE);
        /* http://bugzilla.gnome.org/show_bug.cgi?id=56070 workaround */
        gtk_widget_hide(pctrl.update);
        gtk_widget_show(pctrl.update);
    }
    else {
        gimp_drawable_flush(drawable);
        gimp_drawable_merge_shadow(drawable->drawable_id, TRUE);
        gimp_drawable_update(drawable->drawable_id,
                             ix1, iy1, (ix2 - ix1), (iy2 - iy1));
    }

}

/* move one-channel data into preview or drawable */
static void
commit(GimpDrawable *drawable, gint channel)
{
    gpointer pr;
    GimpPixelRgn dest_rgn;

    if (preview_mode) {
        dest_rgn.x = dest_rgn.y = 0;
        dest_rgn.w = dest_rgn.h = PREVIEW_SIZE;
        dest_rgn.bpp = gdk_pixbuf_get_n_channels(preview);
        dest_rgn.rowstride = gdk_pixbuf_get_rowstride(preview);
        dest_rgn.data = gdk_pixbuf_get_pixels(preview);
        commit_one_tile(&dest_rgn, channel, gradient_p_width);
    }
    else {
        gimp_pixel_rgn_init(&dest_rgn, drawable,
                            ix1, iy1, (ix2 - ix1), (iy2 - iy1),
                            TRUE, TRUE);
        /* iterate by tiles */
        for (pr = gimp_pixel_rgns_register(1, &dest_rgn);
             pr != NULL;
             pr = gimp_pixel_rgns_process(pr)) {
            commit_one_tile(&dest_rgn, channel, GRADIENT_SAMPLE_SIZE);
        }
    }
}

/* move one-channel data into one tile of drawable
   when channel == alpha, just fill it with CHANNEL_MAX_VALUE */
static void
commit_one_tile(GimpPixelRgn *rgn, gint channel, gint gradwidth)
{
    gint x, y;
    guchar *p;
    gfloat *data = pdata;
    gfloat min = data_range.min;
    gfloat max = data_range.max;
    gfloat r;
    gint index;

    if (channel == alpha && pvals.cm != COLORING_METHOD_GRADIENT) {
        for (y = rgn->y; y < rgn->y + rgn->h; y++) {
            for (x = rgn->x; x < rgn->x + rgn->w; x++) {
                p = rgn->data
                    + (y - rgn->y)*rgn->rowstride
                    + (x - rgn->x)*rgn->bpp;
                p[channel] = CHANNEL_MAX_VALUE;
            }
        }
        return;
    }

    if (pvals.cm == COLORING_METHOD_GRADIENT) {
        for (y = rgn->y; y < rgn->y + rgn->h; y++) {
            for (x = rgn->x; x < rgn->x + rgn->w; x++) {
                p = rgn->data
                    + (y - rgn->y)*rgn->rowstride
                    + (x - rgn->x)*rgn->bpp;
                r = data[(y - iy1)*(ix2 - ix1) + x - ix1];
                r = (r - min)/(max - min)*gradwidth;
                index = CLAMP(r, 0, gradwidth-1);
                r = gradient[4*index + channel];
                p[channel] = CHANNEL_MAX_VALUE*r;
            }
        }
    }
    else {
        for (y = rgn->y; y < rgn->y + rgn->h; y++) {
            for (x = rgn->x; x < rgn->x + rgn->w; x++) {
                p = rgn->data + (y - rgn->y)*rgn->rowstride
                    + (x - rgn->x)*rgn->bpp;
                r = data[(y - iy1)*(ix2 - ix1) + x - ix1];
                r = (r - min)/(max - min);
                r = CLAMP(r, 0.0, 1.0);
                p[channel] = CHANNEL_MAX_VALUE*r;
            }
        }
    }
}

static void
add_transparency(gint channel)
{
    static gint check_colors[] = {
        CHANNEL_MAX_VALUE*GIMP_CHECK_DARK,
        CHANNEL_MAX_VALUE*GIMP_CHECK_LIGHT,
    };

    guchar *p;
    gint x, y, i, chc;
    gdouble f;

    for (x = ix1; x < ix2; x++) {
        for (y = iy1; y < iy2; y++) {
            i = (pdata[(x - ix1) + (y - iy1)*(ix2 - ix1)] - data_range.min)/
                (data_range.max - data_range.min)*(gradient_p_width - 1);
            i = CLAMP(i, 0, gradient_p_width-1);
            f = gradient[4*i + 3];
            p = gdk_pixbuf_get_pixels(preview)
                + y*gdk_pixbuf_get_rowstride(preview) + x*bpp + channel;
            chc = (x - ix1)%(2*GIMP_CHECK_SIZE)/GIMP_CHECK_SIZE
                + (y - iy1)%(2*GIMP_CHECK_SIZE)/GIMP_CHECK_SIZE;
            *p = (*p)*f + (1 - f)*check_colors[chc % 2];
        }
    }
}

static void
render_gradient(void)
{
    gint dummy;

    if (pvals.cm != COLORING_METHOD_GRADIENT)
        return;

    if (preview_mode) {
        /* don't regenerate what we already have */
        if (gradient == NULL) {
            gradient_p_width = PREVIEW_SIZE;
            gimp_gradient_get_uniform_samples(gimp_context_get_gradient(),
                                              gradient_p_width,
                                              FALSE,
                                              &dummy, &gradient);
        }
    }
    else {
        /* final rendering, done only once at the end */
        g_free(gradient);
        gimp_gradient_get_uniform_samples(gimp_context_get_gradient(),
                                          GRADIENT_SAMPLE_SIZE,
                                          FALSE,
                                          &dummy, &gradient);
    }
}

/***** Turing Pattern *******************************************************/

static void
turingp_tp(void)
{
    gint i, s, cxres, cyres, nxres, nyres;
    gfloat *pdata_tmp;
    gfloat feature_size, x;
    Stage stages[4];

    s = 0;
    cxres = ix2 - ix1;
    cyres = iy2 - iy1;
    feature_size = pvals.feature_size;
    while (TRUE) {
        stages[s].xres = cxres;
        stages[s].yres = cyres;
        stages[s].h = FEATURE_SIZE_TO_H/feature_size;
        stages[s].iscale = cxres*cyres;
        if (preview_mode)
            stages[s].iter = 300;
        else
            stages[s].iter = 1000;
        s++;

        if (feature_size <= 6.25)
            break;

        x = MAX(feature_size/2, 5.0);
        cxres = MAX(x/feature_size*cxres, 1);
        cyres = MAX(x/feature_size*cyres, 1);
        feature_size = x;
    }

    max_progress = 0;
    for (i = 0; i < s; i++) {
        stages[i].iscale /= stages[s-1].iscale;
        max_progress += stages[i].iter*stages[i].iscale;
    }

    s--;
    cxres = stages[s].xres;
    cyres = stages[s].yres;
    pdata = g_new(gfloat, 4*cxres*cyres);
    for (i = 0; i < 2*cxres*cyres; i++)
        pdata[i] = g_rand_double_range(rng, -0.5, 0.5);

    progress = 0;
    while (TRUE) {
        iterate_tp(cxres, cyres, stages[s].iter, stages[s].h, stages[s].iscale);

        if (!s)
            break;

        s--;
        nxres = stages[s].xres;
        nyres = stages[s].yres;
        pdata_tmp = g_new(gfloat, 4*nxres*nyres);
        scale_up_periodic(cxres, cyres, pdata,
                          nxres, nyres, pdata_tmp);
        scale_up_periodic(cxres, cyres, pdata + cxres*cyres,
                          nxres, nyres, pdata_tmp + nxres*nyres);
        g_free(pdata);
        pdata = pdata_tmp;
        cxres = nxres;
        cyres = nyres;
    }

    data_range = normalize_hist(stages[0].xres*stages[0].yres, pdata);
}

static void
iterate_tp(gint width,
           gint height,
           gint iterations,
           gfloat h,
           gint iscale)
{
    gint i, n;

    n = width*height;

    for (i = 0; i < iterations; i++) {
        do_iter_tp(width, height, pdata, h);
        progress += iscale;
        if (preview_mode) {
            if (!(i % 100)) {
                gtk_progress_bar_set_fraction
                                       (GTK_PROGRESS_BAR(preview_progressbar),
                                        (gdouble)progress/max_progress);
                while (gtk_events_pending())
                    gtk_main_iteration_do(FALSE);
            }
        }
        else {
            if (!(i % PROGRESS_STEP))
                gimp_progress_update(progress/max_progress);
        }
    }

    if (preview_mode) {
        gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(preview_progressbar),
                                      0.0);
        while (gtk_events_pending())
            gtk_main_iteration_do(FALSE);
    }
}

static void
do_iter_tp(gint xres,
           gint yres,
           gfloat *data,
           gfloat h)
{
    gfloat *c0, *c1, *r0, *r1;
    gfloat p = 1.12;
    gfloat q = -1.4;
    gfloat p0 = -1.2;
    gfloat q0 = 0.8;
    gfloat mu0h, mu1h, realdt, rr0, rr1, cr0, cr1;
    gint i, j, ix, ixp, ixm, jp, jm;

    c0 = data;
    c1 = data + xres*yres;
    r0 = data + 2*xres*yres;
    r1 = data + 3*xres*yres;

    mu0h = 0.00001/h/h;
    mu1h = 0.0001/h/h;

    cr0 = cr1 = 0.0;
    for (i = 0; i < yres; i++) {
        ix = i*xres;
        ixp = ((i + 1) % yres)*xres;
        ixm = ((i + yres - 1) % yres)*xres;
        for (j = 0; j < xres; j++) {
            gfloat c0lap, c1lap, cx0, cx1;

            jp = (j + 1) % xres;
            jm = (j + xres - 1) % xres;
            cx0 = c0[ix + j];
            cx1 = c1[ix + j];
            /* mixed rectangular-diagonal laplaceans */
            c0lap = c0[ix + jp] + c0[ix + jm] + c0[ixp + j] + c0[ixm + j]
                    + 0.25*(c0[ixp + jp] + c0[ixp + jm]
                            + c0[ixm + jp] + c0[ixm + jm])
                    - 5.0*cx0;
            c1lap = c1[ix + jp] + c1[ix + jm] + c1[ixp + j] + c1[ixm + j]
                    + 0.25*(c1[ixp + jp] + c1[ixp + jm]
                            + c1[ixm + jp] + c1[ixm + jm])
                    - 5.0*cx1;
            r0[ix + j] = q0*F(cx0) + q*cx1 + mu0h*c0lap;
            r1[ix + j] = p0*F(cx1) + p*cx0 + mu1h*c1lap;

            cr0 += cx0*cx0;
            cr1 += cx1*cx1;
        }
    }

    checker_smooth(xres, yres, r0, r1, &rr0, &rr1);
    rr0 = sqrt(cr0/rr0);
    rr1 = sqrt(cr1/rr1);
    realdt = 0.2*MIN(rr0, rr1);
    for (i = 0; i < xres*yres; i++) {
        c0[i] += realdt*r0[i];
        c1[i] += realdt*r1[i];
    }
}

/***** Conglomerate Rock ****************************************************/

static void
turingp_cr(void)
{
    gint i, xres, yres;

    xres = ix2 - ix1;
    yres = iy2 - iy1;
    max_progress = 322;
    progress = 0;

    pdata = g_new(gfloat, 4*xres*yres);
    for (i = 0; i < 2*xres*yres; i++)
        pdata[i] = g_rand_double_range(rng, -0.5, 0.5);

    iterate_cr(xres, yres, max_progress);
    memcpy(pdata, pdata + xres*yres, xres*yres*sizeof(gfloat));

    data_range = normalize(xres*yres, pdata);
}

static void
iterate_cr(gint width,
           gint height,
           gint iterations)
{
    gint i, n;

    n = width*height;

    for (i = 0; i < iterations; i++) {
        do_iter_cr(width, height, pdata);
        progress++;
        if (preview_mode) {
            if (!(i % PROGRESS_PREVIEW_STEP)) {
                gtk_progress_bar_set_fraction
                                       (GTK_PROGRESS_BAR(preview_progressbar),
                                        progress/max_progress);
                while (gtk_events_pending())
                    gtk_main_iteration_do(FALSE);
            }
        }
        else {
            if (!(i % PROGRESS_STEP))
                gimp_progress_update(progress/max_progress);
        }
    }

    if (preview_mode) {
        gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(preview_progressbar),
                                      0.0);
        while (gtk_events_pending())
            gtk_main_iteration_do(FALSE);
    }
}

static void
do_iter_cr(gint xres,
           gint yres,
           gfloat *data)
{
    static gfloat *c0, *c1, *r0, *r1;
    gfloat h = 1.0/30;
    gfloat p = 3.12;
    gfloat p0 = 0.3;
    gfloat q = -3.4;
    gfloat q0 = 0.8;
    gfloat mu0h, mu1h, realdt, rr0, rr1, cr0, cr1;
    gint i, j, ix, ixp, ixm, jp, jm;

    c0 = data;
    c1 = data + xres*yres;
    r0 = data + 2*xres*yres;
    r1 = data + 3*xres*yres;

    mu0h = 0.00002/h/h;
    mu1h = 0.0002/h/h;

    cr0 = cr1 = 0.0;
    for (i = 0; i < yres; i++) {
        ix = i*xres;
        ixp = ((i + 1) % yres)*xres;
        ixm = ((i + yres - 1) % yres)*xres;
        for (j = 0; j < xres; j++) {
            gdouble c0lap, c1lap, cx0, cx1;

            jp = (j + 1) % xres;
            jm = (j + xres - 1) % xres;
            cx0 = c0[ix + j];
            cx1 = c1[ix + j];
            /* mixed rectangular-diagonal laplaceans */
            c0lap = c0[ix + jp] + c0[ix + jm] + c0[ixp + j] + c0[ixm + j]
                    + 0.25*(c0[ixp + jp] + c0[ixp + jm]
                            + c0[ixm + jp] + c0[ixm + jm])
                    - 5.0*cx0;
            c1lap = c1[ix + jp] + c1[ix + jm] + c1[ixp + j] + c1[ixm + j]
                    + 0.25*(c1[ixp + jp] + c1[ixp + jm]
                            + c1[ixm + jp] + c1[ixm + jm])
                    - 5.0*cx1;
            r0[ix + j] = q0*F(cx0) + q*F(cx1) + mu0h*c0lap;
            r1[ix + j] = p0*F(cx1) + p*F(cx0) + mu1h*c1lap;
            cr0 += cx0*cx0;
            cr1 += cx1*cx1;
        }
    }

    checker_smooth(xres, yres, r0, r1, &rr0, &rr1);
    rr0 = sqrt(cr0/rr0);
    rr1 = sqrt(cr1/rr1);
    realdt = 0.1*MIN(rr0, rr1);
    for (i = 0; i < xres*yres; i++) {
        c0[i] += realdt*r0[i];
        c1[i] += realdt*r1[i];
    }
}

/***** Protoplasm/Cheetah ***************************************************/

static void
turingp_pc(gint channel)
{
    gint i, n, xres, yres;

    xres = ix2 - ix1;
    yres = iy2 - iy1;
    n = xres*yres;
    max_progress = 500;
    progress = 0;

    pdata = g_new(gfloat, (4 + channel)*n);
    for (i = 0; i < 2*n; i++)
        pdata[i] = g_rand_double_range(rng, -0.5, 0.5);

    if (!channel)
        iterate_pc(xres, yres, 3.12, -4.0, 3.1, 1.2, max_progress);
    else {
        iterate_pc(xres, yres, 3.12, -4.0, 3.1, 1.2, max_progress-1);
        memcpy(pdata + 4*n, pdata + n, n*sizeof(gfloat));
        iterate_pc(xres, yres, 3.12, -4.0, 3.1, 1.2, 1);
        for (i = 0; i < n; i++)
            pdata[i] = pdata[n + i] + pdata[4*n + i];
    }

    data_range = normalize_hist(n, pdata);
}

static void
iterate_pc(gint width,
           gint height,
           gfloat p,
           gfloat q,
           gfloat p0,
           gfloat q0,
           gint iterations)
{
    gint i, n;

    n = width*height;

    for (i = 0; i < iterations; i++) {
        do_iter_pc(width, height, pdata, p, q, p0, q0);
        progress++;
        if (preview_mode) {
            if (!(i % PROGRESS_PREVIEW_STEP)) {
                gtk_progress_bar_set_fraction
                                       (GTK_PROGRESS_BAR(preview_progressbar),
                                        progress/max_progress);
                while (gtk_events_pending())
                    gtk_main_iteration_do(FALSE);
            }
        }
        else {
            if (!(i % PROGRESS_STEP))
                gimp_progress_update(progress/max_progress);
        }
    }

    if (preview_mode) {
        gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(preview_progressbar),
                                      0.0);
        while (gtk_events_pending())
            gtk_main_iteration_do(FALSE);
    }
}

static void
do_iter_pc(gint xres,
           gint yres,
           gfloat *data,
           gfloat p,
           gfloat q,
           gfloat p0,
           gfloat q0)
{
    static gfloat *c0, *c1, *r0, *r1;
    gfloat h = 1.0/30;
    gfloat mu0h, mu1h, realdt, rr0, rr1, cr0, cr1;
    gint i, j, ix, ixp, ixm, jp, jm;

    c0 = data;
    c1 = data + xres*yres;
    r0 = data + 2*xres*yres;
    r1 = data + 3*xres*yres;

    mu0h = 0.00002/h/h;
    mu1h = 0.0002/h/h;

    cr0 = cr1 = 0.0;
    for (i = 0; i < yres; i++) {
        ix = i*xres;
        ixp = ((i + 1) % yres)*xres;
        ixm = ((i + yres - 1) % yres)*xres;
        for (j = 0; j < xres; j++) {
            gdouble c0lap, c1lap, cx0, cx1;

            jp = (j + 1) % xres;
            jm = (j + xres - 1) % xres;
            cx0 = c0[ix + j];
            cx1 = c1[ix + j];
            /* mixed rectangular-diagonal laplaceans */
            c0lap = c0[ix + jp] + c0[ix + jm] + c0[ixp + j] + c0[ixm + j]
                    + 0.25*(c0[ixp + jp] + c0[ixp + jm]
                            + c0[ixm + jp] + c0[ixm + jm])
                    - 5.0*cx0;
            c1lap = c1[ix + jp] + c1[ix + jm] + c1[ixp + j] + c1[ixm + j]
                    + 0.25*(c1[ixp + jp] + c1[ixp + jm]
                            + c1[ixm + jp] + c1[ixm + jm])
                    - 5.0*cx1;
            r0[ix + j] = q0*F(cx0) + q*F(cx1) + mu0h*c0lap;
            r1[ix + j] = p0*F(cx1) + p*F(cx0) + mu1h*c1lap;
            cr0 += cx0*cx0;
            cr1 += cx1*cx1;
        }
    }

    checker_smooth(xres, yres, r0, r1, &rr0, &rr1);
    rr0 = sqrt(cr0/rr0);
    rr1 = sqrt(cr1/rr1);
    realdt = 0.1*MIN(rr0, rr1);
    for (i = 0; i < xres*yres; i++) {
        c0[i] += realdt*r0[i];
        c1[i] += realdt*r1[i];
    }
}

/***** Common **************************************************************/

/* smooth derivations to prevent laplacean `checker' instability */
static void
checker_smooth(gint xres, gint yres, gfloat *r0, gfloat *r1,
               gfloat *prr0, gfloat *prr1)
{
    gint i, j, ix, ixp, ixm, jp, jm;
    gfloat rr0, rr1;

    rr0 = rr1 = 0.0;
    for (i = 0; i < yres; i++) {
        ix = i*xres;
        ixp = ((i + 1) % yres)*xres;
        ixm = ((i + yres - 1) % yres)*xres;
        for (j = 0; j < xres; j++) {
            jp = (j + 1) % xres;
            jm = (j + xres - 1) % xres;

            r0[ix + j] += 0.125*(r0[ixp + j] + r0[ixm + j]
                                 + r0[ix + jp] + r0[ix + jm]);
            r0[ix + j] /= 1.5;
            rr0 += r0[ix + j]*r0[ix + j];

            r1[ix + j] += 0.125*(r1[ixp + j] + r1[ixm + j]
                                 + r1[ix + jp] + r1[ix + jm]);
            r1[ix + j] /= 1.5;
            rr1 += r1[ix + j]*r1[ix + j];
        }
    }

    *prr0 = rr0;
    *prr1 = rr1;
}

/* scale up a field with periodic boundary conditions.  this is somewhat
 * different from normal scaling that would `leak' a strip of pixels on
 * the edge.  we do it by interpreting pixels values as being in lower left
 * corner of the pixel and using periodic boundary conditions on the upper
 * right ones.  */
static void
scale_up_periodic(gint sxres, gint syres, gfloat *src,
                  gint dxres, gint dyres, gfloat *dst)
{
    gint i, j, ix0, ix1, j0, j1;
    gfloat *drow;
    gfloat x, y;

    for (i = 0; i < dyres; i++) {
        drow = dst + i*dxres;
        y = (gfloat)i/dyres*syres;
        ix0 = (gint)floor(y);
        ix1 = (ix0 + 1) % syres;
        y -= ix0;
        ix0 *= sxres;
        ix1 *= sxres;

        for (j = 0; j < dxres; j++) {
            x = (gfloat)j/dxres*sxres;
            j0 = (gint)floor(x);
            j1 = (j0 + 1) % sxres;
            x -= j0;

            drow[j] = (1.0 - x)*((1.0 - y)*src[ix0 + j0] + y*src[ix1 + j0])
                      + x*((1.0 - y)*src[ix0 + j1] + y*src[ix1 + j1]);
        }
    }
}

static DataRange
normalize(gint n, gfloat *data)
{
    DataRange dr;
    gfloat avg, rms, max, min;
    gint i;

    min = G_MAXFLOAT;
    max = -G_MAXFLOAT;
    for (i = 0; i < n; i++) {
        if (data[i] < min)
            min = data[i];
        if (data[i] > max)
            max = data[i];
    }

    avg = rms = 0.0;
    for (i = 0; i < n; i++)
        avg += data[i];
    avg /= n;
    for (i = 0; i < n; i++)
        rms += (data[i] - avg)*(data[i] - avg);
    rms = sqrt(rms/n);

    dr.min = MAX(avg - 3.0*rms, min);
    dr.max = MIN(avg + 3.0*rms, max);
    return dr;
}

static DataRange
normalize_hist(gint n, gfloat *data)
{
    DataRange dr;
    gint *count;
    gint i, sum, nhist;
    gfloat min, max;

    min = G_MAXFLOAT;
    max = -G_MAXFLOAT;
    for (i = 0; i < n; i++) {
        if (data[i] < min)
            min = data[i];
        if (data[i] > max)
            max = data[i];
    }

    nhist = 3*sqrt(n);
    count = g_new0(gint, nhist + 1);
    for (i = 0; i < n; i++)
        count[(gint)((data[i] - min)/(max - min)*nhist)]++;

    sum = 0;
    for (i = 0; sum < n/100; i++)
        sum += count[i];
    dr.min = (gfloat)i/nhist*(max - min) + min;

    sum = 0;
    for (i = nhist-1; sum < n/100; i--)
        sum += count[i];
    dr.max = (gfloat)(i+1)/nhist*(max - min) + min;

    return dr;
}

